import stesting

def registerTestSetup():
    from zope.interface import classImplements
    from schooltool.testing import registry

    def haveCalendar():
        from schooltool.course import section
        from schooltool.app.interfaces import IHaveCalendar
        if not IHaveCalendar.implementedBy(section.Section):
            classImplements(section.Section, IHaveCalendar)
    registry.register('CalendarComponents', haveCalendar)

    def bookResources():
        from schooltool.course import section
        from schooltool.resource.interfaces import IBookResources
        if not IBookResources.implementedBy(section.Section):
            classImplements(section.Section, IBookResources)

    registry.register('ResourceComponents', bookResources)

registerTestSetup()
del registerTestSetup




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Customisation for SchoolTool security policy.
"""

from zope.interface import implements
from zope.annotation.interfaces import IAnnotations
from zope.component import subscribers
from persistent import Persistent
from persistent.dict import PersistentDict
from schooltool.securitypolicy.interfaces import IAccessControlCustomisations
from schooltool.securitypolicy.interfaces import IAccessControlSetting


class AccessControlCustomisations(Persistent):
    implements(IAccessControlCustomisations)

    def __init__(self):
        self._settings = PersistentDict()

    def getSetting(self, key):
        for setting in self:
            if setting.key == key:
                return setting
        else:
            raise KeyError("there is no AccessControlSetting"
                           " associated with this key.")

    def get(self, key):
        return self._settings.get(key, self.getSetting(key).default)

    def set(self, key, value):
        if self.getSetting(key):
            self._settings[key] = value

    def __iter__(self):
        settings = subscribers([None], IAccessControlSetting)
        return iter(settings)


def getAccessControlCustomisations(app):
    """Adapt a SchoolToolApplication to IAccessControlCustomisations."""

    annotations = IAnnotations(app)
    key = 'schooltool.securitypolicy.AccessControlCustomisations'
    try:
        return annotations[key]
    except KeyError:
        annotations[key] = AccessControlCustomisations()
        annotations[key].__parent__ = app
        return annotations[key]



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Demographics fields and storage
"""
from persistent.dict import PersistentDict
from persistent import Persistent

from zope.schema._field import Choice
from zope.schema._field import Date
from zope.schema import TextLine, Bool
from zope.schema import Int
from zope.location.location import Location
from zope.location.location import locate
from zope.html.field import HtmlFragment
from zope.interface import Interface
from zope.interface import implements
from zope.interface import implementer
from zope.component import adapts
from zope.component import adapter
from zope.container.btree import BTreeContainer
from zope.container.ordered import OrderedContainer
from zope.lifecycleevent.interfaces import IObjectRemovedEvent
from zope.schema.vocabulary import SimpleVocabulary
from zope.schema.vocabulary import SimpleTerm

import z3c.form.field

from schooltool.app.app import InitBase, StartUpBase
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.basicperson.interfaces import IEnumFieldDescription
from schooltool.basicperson.interfaces import IIntFieldDescription
from schooltool.basicperson.interfaces import IDemographicsFields
from schooltool.basicperson.interfaces import IBasicPerson
from schooltool.basicperson.interfaces import IDemographics
from schooltool.basicperson.interfaces import IFieldDescription
from schooltool.common import SchoolToolMessage as _


class IDemographicsForm(Interface):
    """Interface for fields that are supposed to get stored in person demographics."""


class PersonDemographicsDataContainer(BTreeContainer):
    """Storage for demographics information for all persons."""


class InvalidKeyError(Exception):
    """Key is not in demographics fields."""


class PersonDemographicsData(PersistentDict):
    """Storage for demographics information for a person."""

    def isValidKey(self, key):
        app = ISchoolToolApplication(None)
        demographics_fields = IDemographicsFields(app)
        return key in demographics_fields

    def __repr__(self):
        return '%s: %s' % (
            object.__repr__(self),
            super(PersonDemographicsData, self).__repr__())

    def __setitem__(self, key, v):
        if not self.isValidKey(key):
            raise InvalidKeyError(key)
        super(PersonDemographicsData, self).__setitem__(key, v)

    def __getitem__(self, key):
        if key not in self and self.isValidKey(key):
            self[key] = None
        return super(PersonDemographicsData, self).__getitem__(key)


@adapter(IBasicPerson)
@implementer(IDemographics)
def getPersonDemographics(person):
    app = ISchoolToolApplication(None)
    pdc = app['schooltool.basicperson.demographics_data']
    demographics = pdc.get(person.username, None)
    if demographics is None:
        pdc[person.username] = demographics = PersonDemographicsData()
    return demographics


@adapter(IBasicPerson, IObjectRemovedEvent)
def removePersonDemographicsSubscriber(person, event):
    app = ISchoolToolApplication(None)
    pdc = app['schooltool.basicperson.demographics_data']
    demographics = pdc.get(person.username, None)
    if demographics is not None:
        del pdc[person.username]


class DemographicsFormAdapter(object):
    implements(IDemographicsForm)
    adapts(IBasicPerson)

    def __init__(self, context):
        self.__dict__['context'] = context
        self.__dict__['demographics'] = IDemographics(self.context)

    def __setattr__(self, name, value):
        self.demographics[name] = value

    def __getattr__(self, name):
        return self.demographics.get(name, None)


class DemographicsFields(OrderedContainer):
    implements(IDemographicsFields)

    def filter_key(self, key):
        """Return the subset of fields whose limited_keys list is either
           empty, or it contains the key passed."""
        result = []
        for field in self.values():
            if not field.limit_keys or key in field.limit_keys:
                result.append(field)
        return result

    def filter_keys(self, keys):
        """Return the subset of fields whose limited_keys list is either
           empty, or it contains one of the keys passed."""
        result = []
        for field in self.values():
            limit_keys = field.limit_keys
            if not limit_keys or [key for key in keys if key in limit_keys]:
                result.append(field)
        return result


def setUpDefaultDemographics(app):
    dfs = DemographicsFields()
    app['schooltool.basicperson.demographics_fields'] = dfs
    locate(dfs, app, 'schooltool.basicperson.demographics_fields')
    dfs['ID'] = TextFieldDescription('ID', _('ID'))
    dfs['ethnicity'] = EnumFieldDescription('ethnicity', _('Ethnicity'))
    dfs['ethnicity'].items = [_('American Indian or Alaska Native'),
                              _('Asian'),
                              _('Black or African American'),
                              _('Native Hawaiian or Other Pacific Islander'),
                              _('White')]
    dfs['language'] = TextFieldDescription('language', _('Language'))
    dfs['placeofbirth'] = TextFieldDescription('placeofbirth', _('Place of birth'))
    dfs['citizenship'] = TextFieldDescription('citizenship', _('Citizenship'))


class DemographicsAppStartup(StartUpBase):
    def __call__(self):
        if 'schooltool.basicperson.demographics_fields' not in self.app:
            setUpDefaultDemographics(self.app)
        if 'schooltool.basicperson.demographics_data' not in self.app:
            self.app['schooltool.basicperson.demographics_data'] = PersonDemographicsDataContainer()


class DemographicsInit(InitBase):
    def __call__(self):
        setUpDefaultDemographics(self.app)
        self.app['schooltool.basicperson.demographics_data'] = PersonDemographicsDataContainer()


@implementer(IDemographicsFields)
@adapter(ISchoolToolApplication)
def getDemographicsFields(app):
    return app['schooltool.basicperson.demographics_fields']


class FieldDescription(Persistent, Location):
    implements(IFieldDescription)
    limit_keys = []

    def __init__(self, name, title, required=False, limit_keys=[]):
        self.name, self.title, self.required, self.limit_keys = (name,
            title, required, limit_keys)

    def setUpField(self, form_field):
        form_field.required = self.required
        form_field.__name__ = str(self.__name__)
        form_field.interface = IDemographicsForm
        return z3c.form.field.Fields(form_field)


# XXX: IMHO all IDNA conversions should be replaced by punycode.
#      64 max length limitation simply breaks things too often.
class IDNAVocabulary(SimpleVocabulary):

    def createTerm(cls, *args):
        """Create a single term from data.

        Encode the value using idna encoding so it would look sane in
        the form if it's ascii, but still work if it uses unicode.
        """
        value = args[0]
        token = value.encode('idna')
        title = value
        return SimpleTerm(value, token, title)
    createTerm = classmethod(createTerm)


class EnumFieldDescription(FieldDescription):
    implements(IEnumFieldDescription)

    items = []

    def makeField(self):
        return self.setUpField(Choice(
                title=unicode(self.title),
                vocabulary=IDNAVocabulary.fromValues(self.items)
                ))


class DateFieldDescription(FieldDescription):

    def makeField(self):
        return self.setUpField(Date(title=unicode(self.title)))


class TextFieldDescription(FieldDescription):

    def makeField(self):
        return self.setUpField(TextLine(title=unicode(self.title)))


class BoolFieldDescription(FieldDescription):

    def makeField(self):
        return self.setUpField(Bool(title=unicode(self.title)))


class DescriptionFieldDescription(FieldDescription):

    def makeField(self):
        return self.setUpField(HtmlFragment(title=unicode(self.title)))


class IntFieldDescription(FieldDescription):

    implements(IIntFieldDescription)

    min_value = None
    max_value = None

    def makeField(self):
        return self.setUpField(Int(title=unicode(self.title),
                                   min=self.min_value,
                                   max=self.max_value))



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
csv importing.
"""
from zope.traversing.browser.absoluteurl import absoluteURL

from schooltool.app.browser.csvimport import BaseCSVImporter
from schooltool.app.browser.csvimport import BaseCSVImportView
from schooltool.app.browser.csvimport import FlourishBaseCSVImportView
from schooltool.resource.resource import Resource, Location

from schooltool.common import SchoolToolMessage as _


def resourceFactory(self, title=u"", description=u"", isLocation=False):
    if isLocation:
        return Location(title=title, description=description)
    else:
        return Resource(title=title, description=description)


class ResourceCSVImporter(BaseCSVImporter):
    """Resource CSV Importer"""

    factory = resourceFactory

    def createAndAdd(self, data, dry_run=True):
        """Create objects and add them to the container."""

        if len(data) < 1:
            self.errors.fields.append(_('Insufficient data provided.'))
            return

        if not data[0]:
            self.errors.fields.append(_('Titles may not be empty'))
            return

        if len(data) > 1:
            description = data[1]
        else:
            description = ''

        isLocation =  len(data) > 2

        obj = self.factory(title=data[0], description=description,
                           isLocation=isLocation)
        name = self.chooser.chooseName('', obj)

        if not dry_run:
            self.container[name] = obj


class ResourceCSVImportView(BaseCSVImportView):
    """View for Resource CSV importer."""

    importer_class = ResourceCSVImporter


class FlourishResourceCSVImportView(FlourishBaseCSVImportView):

    importer_class = ResourceCSVImporter

    def nextURL(self):
        return absoluteURL(self.context, self.request)



#
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Contact relationship views.
"""
from zope.security.proxy import removeSecurityProxy
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.publisher.browser import BrowserView
from zope.catalog.interfaces import ICatalog
from zope.intid.interfaces import IIntIds
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.traversing.browser.absoluteurl import absoluteURL

from zc.table.column import GetterColumn

from schooltool.relationship.relationship import IRelationshipLinks
from schooltool.basicperson.interfaces import IBasicPerson
from schooltool.app.browser.app import RelationshipAddTableMixin
from schooltool.app.browser.app import RelationshipRemoveTableMixin
from schooltool.app.browser.app import AddAllResultsButton
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.common.inlinept import InheritTemplate
from schooltool.contact.interfaces import IContactable
from schooltool.contact.interfaces import IContact
from schooltool.contact.interfaces import IContactContainer
from schooltool.contact.interfaces import IUniqueFormKey
from schooltool.contact.contact import URIPerson, URIContact
from schooltool.contact.contact import URIContactRelationship
from schooltool.contact.contact import ContactPersonInfo
from schooltool.contact.browser.contact import ContactTable
from schooltool.contact.browser.contact import contact_table_columns
from schooltool.table.table import CheckboxColumn
from schooltool.table.table import label_cell_formatter_factory
from schooltool.table.interfaces import ITableFormatter
from schooltool.app.browser.app import EditRelationships
from schooltool.skin.flourish.page import Page

from schooltool.common import SchoolToolMessage as _


def get_relationship_title(person, contact):
    try:
        links = IRelationshipLinks(person)
        link = links.find(
            URIPerson, contact, URIContact, URIContactRelationship)
    except ValueError:
        return u''
    return link.extra_info.getRelationshipTitle()


def make_relationship_column_getter(person=None):
    def format_item(item, formatter):
        if person is None:
            return u''
        item = removeSecurityProxy(item)
        return get_relationship_title(person, item)
    return format_item


def assigned_contacts_columns(person=None):
    default_columns = contact_table_columns()
    relationship = GetterColumn(
        name='relationship',
        title=_(u"Relationship"),
        getter=make_relationship_column_getter(person))
    return default_columns + [relationship]


class ContactManagementView(BrowserView):
    """View class for adding/removing contacts to/from a person."""

    __used_for__ = IBasicPerson

    __call__ = ViewPageTemplateFile('templates/manage_contacts.pt')
    current_title = _("Assigned Contacts")
    available_title = _("Assign existing contact")

    @property
    def title(self):
        return _("Manage contacts of ${person}",
            mapping={'person': self.context.title})

    def add(self, item):
        """Add an item to the list of selected items."""
        collection = removeSecurityProxy(self.getCollection())
        info = ContactPersonInfo()
        info.__parent__ = removeSecurityProxy(self.context)
        collection.add(item, info)

    def remove(self, item):
        """Remove an item from selected items."""
        collection = removeSecurityProxy(self.getCollection())
        collection.remove(item)

    def getCollection(self):
        return IContactable(removeSecurityProxy(self.context)).contacts

    def getSelectedItemIds(self):
        int_ids = getUtility(IIntIds)
        return [int_ids.getId(item) for item in self.getCollection()]

    def getItemContainer(self):
        return IContactContainer(ISchoolToolApplication(None))

    def getCatalog(self):
        return ICatalog(self.getItemContainer())

    def getAvailableItemIds(self):
        selected = self.getSelectedItemIds()
        catalog = self.getCatalog()
        return [intid for intid in catalog.extent
                if intid not in selected]

    def getOmmitedItemIds(self):
        int_ids = getUtility(IIntIds)
        self_contact = IContact(self.context)
        return self.getSelectedItemIds() + [int_ids.getId(self_contact)]

    def createTableFormatter(self, **kwargs):
        prefix = kwargs['prefix']
        container = self.getItemContainer()
        formatter = getMultiAdapter((container, self.request),
                                    ITableFormatter)
        columns_before = [CheckboxColumn(
            prefix=prefix, title="", id_getter=IUniqueFormKey)]
        formatters = [label_cell_formatter_factory(prefix, IUniqueFormKey)]
        formatter.setUp(formatters=formatters,
                        columns_before=columns_before,
                        **kwargs)
        return formatter

    def setUpTables(self):
        int_ids = getUtility(IIntIds)
        self.available_table = self.createTableFormatter(
            ommit=[int_ids.getObject(intid)
                   for intid in self.getOmmitedItemIds()],
            prefix="add_item")

        self.selected_table = self.createTableFormatter(
            filter=lambda l: l,
            items=[int_ids.getObject(intid)
                   for intid in self.getSelectedItemIds()],
            columns=assigned_contacts_columns(self.context),
            prefix="remove_item",
            batch_size=0)

    def update(self):
        context_url = absoluteURL(self.context, self.request)
        catalog = self.getCatalog()
        int_ids = getUtility(IIntIds)
        index = catalog['form_keys']
        if 'ADD_ITEMS' in self.request:
            for intid in self.getAvailableItemIds():
                key = 'add_item.%s' % index.documents_to_values[intid]
                if key in self.request:
                    self.add(IContact(int_ids.getObject(intid)))
        elif 'REMOVE_ITEMS' in self.request:
            for intid in self.getSelectedItemIds():
                key = 'remove_item.%s' % index.documents_to_values[intid]
                if key in self.request:
                    self.remove(int_ids.getObject(intid))
        elif 'CANCEL' in self.request:
            self.request.response.redirect(context_url)

        self.setUpTables()


class EditContactRelationships(EditRelationships):
    pass


class ContactAddRelationshipTable(RelationshipAddTableMixin, ContactTable):

    def submitItems(self):
        int_ids = getUtility(IIntIds)
        catalog = self.view.getCatalog()
        index = catalog['form_keys']
        for intid in self.view.getAvailableItemIds():
            key = '%s.%s' % (self.button_prefix, index.documents_to_values[intid])
            if key in self.request:
                self.view.add(IContact(int_ids.getObject(intid)))


class ContactRemoveRelationshipTable(RelationshipRemoveTableMixin, ContactTable):

    def submitItems(self):
        int_ids = getUtility(IIntIds)
        catalog = self.view.getCatalog()
        index = catalog['form_keys']
        for intid in self.view.getSelectedItemIds():
            key = '%s.%s' % (self.button_prefix, index.documents_to_values[intid])
            if key in self.request:
                self.view.remove(IContact(int_ids.getObject(intid)))


class AddAllContactResultsButton(AddAllResultsButton):

    def addSearchResults(self):
        if (self.button_name not in self.request or
            self.token_key not in self.request):
            return False
        add_ids = self.request[self.token_key]
        if not isinstance(add_ids, list):
            add_ids = [add_ids]
        changed = False
        relationship_view = self.manager.view
        int_ids = getUtility(IIntIds)
        for item_id in relationship_view.getAvailableItemIds():
            item = int_ids.getObject(item_id)
            if relationship_view.getKey(item) in add_ids:
                relationship_view.add(removeSecurityProxy(item))
                changed = True
        return changed


class FlourishContactManagementView(EditContactRelationships):

    page_template = InheritTemplate(Page.page_template)

    current_title = _('Current contacts')
    available_title = _('Available contacts')

    def add(self, item):
        """Add an item to the list of selected items."""
        collection = removeSecurityProxy(self.getCollection())
        info = ContactPersonInfo()
        info.__parent__ = removeSecurityProxy(self.context)
        collection.add(item, info)

    def remove(self, item):
        """Remove an item from selected items."""
        collection = removeSecurityProxy(self.getCollection())
        collection.remove(item)

    def getCollection(self):
        return IContactable(removeSecurityProxy(self.context)).contacts

    def getSelectedItemIds(self):
        int_ids = getUtility(IIntIds)
        return [int_ids.getId(item) for item in self.getCollection()]

    def getAvailableItemsContainer(self):
        return IContactContainer(ISchoolToolApplication(None))

    def getCatalog(self):
        return ICatalog(self.getAvailableItemsContainer())

    def getAvailableItemIds(self):
        selected = self.getSelectedItemIds()
        catalog = self.getCatalog()
        return [intid for intid in catalog.extent
                if intid not in selected]

    def getOmmitedItemIds(self):
        int_ids = getUtility(IIntIds)
        self_contact = IContact(self.context)
        return self.getSelectedItemIds() + [int_ids.getId(self_contact)]

    def getKey(self, item):
        return IUniqueFormKey(item)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool person name generator
"""
import os.path
from schooltool.sampledata import PortableRandom


class NameGenerator(object):
    """Person name generator

    Generates random full names.  Names can recur.
    """

    def __init__(self, seed):
        self.random = PortableRandom(seed)
        self.first_names = self._readLines('first_names.txt')
        self.last_names = self._readLines('last_names.txt')

    def _readLines(self, filename):
        """Read in lines from file

        Filename is relative to the module.
        Returned lines are stripped.
        """
        fullpath = os.path.join(os.path.dirname(__file__), filename)
        lines = file(fullpath).readlines()
        return [line.strip() for line in lines]

    def generate(self):
        """Generates name information, returns tuple of
        first name, last name, full name
        """
        first_name = self.random.choice(self.first_names)
        last_name = self.random.choice(self.last_names)
        full_name = "%s %s" % (first_name, last_name)
        return first_name, last_name, full_name




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import zope.schema
import zope.file.interfaces
from zope.publisher.interfaces.browser import IBrowserPage

from schooltool.report.interfaces import IReportMessage
from schooltool.task.interfaces import IProgressMessage
from schooltool.task.interfaces import IRemoteTask

from schooltool.common import SchoolToolMessage as _


class IImporterTask(IRemoteTask):

    xls_file = zope.schema.Object(
        title=_("XLS File"),
        schema=zope.file.interfaces.IFile)


class IXLSExportView(IBrowserPage):

    filename = zope.schema.TextLine(
        title=u"XLS file name", required=False)

    render_invariant = zope.schema.Bool(
        title=u"Render invariant",
        description=u"Render without influence form environment, like current time.",
        required=False)

    render_debug = zope.schema.Bool(
        title=u"Render debug",
        description=u"Render with debug information.",
        required=False)


class IXLSProgressMessage(IProgressMessage, IReportMessage):
    pass



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.app.cal.
"""

import unittest
import doctest
from textwrap import dedent
from datetime import datetime, timedelta

from zope.interface import implements
from zope.interface.verify import verifyObject
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.app.testing import setup

from schooltool.app.interfaces import IHaveCalendar
from schooltool.testing import setup as sbsetup


class ResourceStub(object):
    """Stub resource.

    For the purposes of resource booking, a resource is anything that
    has a 'calendar' attribute.
    """
    implements(IHaveCalendar, IAttributeAnnotatable)


def doctest_CalendarEvent():
    r"""Tests for CalendarEvent.

    CalendarEvents are almost like SimpleCalendarEvents, the main difference
    is that CalendarEvents are mutable and IContained.

        >>> from schooltool.app.cal import CalendarEvent

        >>> event = CalendarEvent(datetime(2005, 2, 7, 16, 24),
        ...                       timedelta(hours=3),
        ...                       "A sample event",
        ...                       unique_id="*the* event")

    The event implements ISchoolToolCalendarEvent:

        >>> from schooltool.app.interfaces import ISchoolToolCalendarEvent
        >>> verifyObject(ISchoolToolCalendarEvent, event)
        True

    It has a name, which is equal to base64 of the unique id, and can
    have a parent:

        >>> event.__name__
        'KnRoZSogZXZlbnQ='
        >>> event.__parent__ is None
        True

    The calendar in which the event resides is referenced in __parent__,
    but you should just use adaptation to ICalendar:

    XXX mg: I have second thoughts about using adaptation to navigate.

        >>> event.__parent__ = object()

        >>> from schooltool.calendar.interfaces import ICalendar
        >>> ICalendar(event) is event.__parent__
        True

    As CalendarEvents are mutable, you can modify attributes at will:

        >>> event.dtend = timedelta(hours=1)
        >>> event.dtend
        datetime.timedelta(0, 3600)


    Owner of the event is the __parent__ of its calendar:

        >>> class CalStub(object):
        ...     __parent__ = "Owner"
        >>> event.__parent__ = CalStub()
        >>> event.owner
        'Owner'

    If event has no calendar it is assumed that it has no owner too:

        >>> event.__parent__ = None
        >>> event.owner is None
        True

    It is very unwise to touch the __name__ or unique_id of events.
    TODO: enforce this restriction.

    """


def doctest_CalendarEvent_resource_booking():
    """Resource booking tests for CalendarEvent.

        >>> setup.setUpAnnotations()
        >>> sbsetup.setUpCalendaring()

    All calendar events have a `resources` argument, but initially it is empty.

        >>> from schooltool.app.interfaces import ISchoolToolCalendar
        >>> from schooltool.app.cal import Calendar, CalendarEvent
        >>> c = Calendar(None)
        >>> e = CalendarEvent(datetime(2005, 2, 28), None, '')
        >>> c.addEvent(e)
        >>> e.resources
        ()

    It is also read-only.

        >>> e.resources = []
        Traceback (most recent call last):
          ...
        AttributeError: can't set attribute

    CalendarEvent.bookResource adds the resource to its resources list and
    adds the event into the resource's calendar.

        >>> r = ResourceStub()
        >>> e.bookResource(r)
        >>> e.resources
        (<...ResourceStub object at...>,)
        >>> e in ISchoolToolCalendar(r)
        True

    CalendarEvent.unbookResource undoes this

        >>> e.unbookResource(r)
        >>> e.resources
        ()
        >>> e in ISchoolToolCalendar(r)
        False

    When you create a CalendarEvent, you can pass a list of resources
    to the constructor.  This is equivalent to creating a CalendarEvent
    without the resources argument, and then subsequently calling bookResource
    for each resource.

        >>> a = ResourceStub()
        >>> b = ResourceStub()
        >>> e = CalendarEvent(datetime(2005, 2, 28), None, 'title', resources=[a, b])
        >>> c.addEvent(e)
        >>> a in e.resources and b in e.resources
        True
        >>> e in ISchoolToolCalendar(a) and e in ISchoolToolCalendar(b)
        True
        >>> e.__parent__ is c
        True

    You cannot book the same resource twice:

        >>> e.bookResource(a)
        Traceback (most recent call last):
          ...
        ValueError: resource already booked

    You cannot unbook a resource that is not booked

        >>> e.unbookResource(r)
        Traceback (most recent call last):
          ...
        ValueError: resource not booked

    """


def doctest_Calendar():
    r"""Tests for Calendar.

    Let's create a Calendar:

        >>> from schooltool.app.cal import Calendar
        >>> cal = Calendar(None)
        >>> cal.__name__
        'calendar'

    The calendar should be an ILocation and it should implement IEditCalendar.

        >>> from schooltool.calendar.interfaces import IEditCalendar
        >>> from zope.location.interfaces import ILocation
        >>> verifyObject(IEditCalendar, cal)
        True
        >>> verifyObject(ILocation, cal)
        True

    A quick look at the empty calendar:

        >>> len(cal)
        0
        >>> list(cal)
        []

    We can add events by using addEvent():

        >>> from schooltool.app.cal import CalendarEvent
        >>> event = CalendarEvent(datetime(2005, 2, 28), None, 'Example 1')

        >>> cal.addEvent(event)
        >>> len(cal)
        1
        >>> list(cal) == [event]
        True

    Added events acquire a parent:

        >>> event.__parent__ is cal
        True

    You should not try to add the same event to a different calendar:

        >>> cal2 = Calendar(None)
        >>> cal2.addEvent(event)
        Traceback (most recent call last):
        ...
        ValueError: Event already belongs to a calendar

    There are exceptions to the previous rule: resource booking adds the same
    event to several calendars simultaneously.

    Let's add a few more events:

        >>> event2 = CalendarEvent(datetime(2005, 2, 28), None, 'Example 2')
        >>> cal.addEvent(event2)
        >>> event3 = CalendarEvent(datetime(2005, 2, 28), None, 'Example 3')
        >>> cal.addEvent(event3)

    You can't, however, add multiple events with the same unique_id

        >>> event3a = CalendarEvent(datetime(2005, 2, 28), None, 'Example 3',
        ...                         unique_id=event3.unique_id)
        >>> cal.addEvent(event3a)
        Traceback (most recent call last):
          ...
        ValueError: an event with this unique_id already exists

    You can iterate through a calendar:

        >>> len(cal)
        3
        >>> titles = [ev.title for ev in cal]
        >>> titles.sort()
        >>> titles
        ['Example 1', 'Example 2', 'Example 3']

    Events can be retrieved by their unique id through find():

        >>> cal.find(event2.unique_id) == event2
        True

        >>> cal.find('nonexistent')
        Traceback (most recent call last):
        ...
        KeyError: 'nonexistent'

    You can remove an event by calling removeEvent()

        >>> cal.removeEvent(event)

    When an event is removed, its __parent__ becomes None

        >>> event.__parent__ is None
        True

    All events can be removed from a calendar by using clear():

        >>> cal.clear()
        >>> list(cal)
        []

    It also clears __parent__ attributes

        >>> event3.__parent__ is None
        True

    By the way, you can specify the calendar's owner in the constructor:

        >>> class PersonStub:
        ...     title = 'A person'
        >>> parent = PersonStub()
        >>> cal2 = Calendar(parent)
        >>> cal2.__parent__ is parent
        True

    The calendar's `title` is its owner's title:

        >>> cal2.title
        'A person'

    We will trust that `expand` inherited from CalendarMixin has been unit
    tested.
    """


def doctest_Calendar_addEvent_resource_booking():
    """Tests for Calendar.addEvent.

        >>> setup.setUpAnnotations()
        >>> sbsetup.setUpCalendaring()

    For the purposes of resource booking, a resource is anything that
    has a 'calendar' attribute.

        >>> resource = ResourceStub()

    When you add an event to a calendar, the calendar's addEvent sets the
    event's __parent__ to itself, or complains loudly if __parent__ already
    set.  Thus normally you cannot add the same calendar event into two
    different calendars.

        >>> from schooltool.app.cal import Calendar, CalendarEvent
        >>> cal = Calendar(None)
        >>> e = CalendarEvent(datetime(2005, 2, 28), None, '')
        >>> cal.addEvent(e)
        >>> e.__parent__ is cal
        True

        >>> from schooltool.app.interfaces import ISchoolToolCalendar
        >>> ISchoolToolCalendar(resource).addEvent(e)
        Traceback (most recent call last):
          ...
        ValueError: Event already belongs to a calendar

    However, there is an exception to the previous rule: you can add an event
    to a resource's calendar, if the resource is listed in event.resources.

        >>> e._resources = (resource, )
        >>> resource in e.resources
        True

        >>> calendar = ISchoolToolCalendar(resource)
        >>> calendar.addEvent(e)
        >>> e in calendar
        True

    Event's __parent__ is not changed in this case.

        >>> e.__parent__ is cal
        True

    """


def doctest_WriteCalendar():
    r"""Tests for WriteCalendar.

    WriteCalendar is an adapter for calendars into IWriteFile

        >>> from schooltool.app.cal import Calendar, WriteCalendar
        >>> cal = Calendar(None)
        >>> writer = WriteCalendar(cal)

        >>> from schooltool.app.interfaces import IWriteCalendar
        >>> verifyObject(IWriteCalendar, writer)
        True

    Let's check that WriteCalendar is hooked up to the iCalendar parser:

        >>> writer.write(dedent('''\
        ... BEGIN:VCALENDAR
        ... VERSION:2.0
        ... METHOD:PUBLISH
        ... BEGIN:VEVENT
        ... UID:956630271
        ... SUMMARY:Christmas Day
        ... CLASS:PUBLIC
        ... DTSTART;VALUE=DATE:20030501
        ... DTEND;VALUE=DATE:20031226
        ... DTSTAMP:20020430T114937Z
        ... END:VEVENT
        ... END:VCALENDAR
        ... '''))

        >>> len(list(cal))
        1
        >>> event = cal.find('956630271')
        >>> event.title
        u'Christmas Day'

    As iCalendar is a bit unwieldy for testing, we will plant a hook:

        >>> from schooltool.calendar.simple import SimpleCalendarEvent
        >>> def event(unique_id, title, day=22, duration=3, **kwargs):
        ...     return SimpleCalendarEvent(title=title,
        ...                                dtstart=datetime(2005, 2, day),
        ...                                duration=duration,
        ...                                unique_id=unique_id)
        >>> evts = []
        >>> writer.read_icalendar = lambda ignored, ignored_too: evts

    OK, now we can do a trivial test with two events:

        >>> evts = [event('e1', 'Event I'), event('e2', 'Event II')]
        >>> writer.write(None)

        >>> len(list(cal))
        2
        >>> cal.find('e1').title
        'Event I'
        >>> cal.find('e2').title
        'Event II'

    """


def doctest_WriteCalendar_sophisticated():
    r"""Tests for WriteCalendar.

    We start with three calendar events.

        >>> from datetime import datetime, timedelta
        >>> from schooltool.app.cal import Calendar, CalendarEvent
        >>> cal = Calendar(None)
        >>> event1 = CalendarEvent(title='Play Doom',
        ...                        dtstart=datetime(2005, 2, 28),
        ...                        duration=timedelta(hours=2),
        ...                        unique_id=u'id1')
        >>> cal.addEvent(event1)
        >>> event2 = CalendarEvent(title='Play chess',
        ...                        dtstart=datetime(2005, 3, 1),
        ...                        duration=timedelta(hours=2),
        ...                        unique_id=u'id2')
        >>> cal.addEvent(event2)
        >>> event3 = CalendarEvent(title='Sleep',
        ...                        dtstart=datetime(2005, 3, 2),
        ...                        duration=timedelta(hours=2),
        ...                        unique_id=u'id3')
        >>> cal.addEvent(event3)

    We will now upload a new calendar that does not change event1, removes
    event2, modifies event3 and adds a new event, event4.

        >>> from schooltool.app.cal import WriteCalendar
        >>> writer = WriteCalendar(cal)
        >>> writer.write(dedent('''\
        ... BEGIN:VCALENDAR
        ... VERSION:2.0
        ... BEGIN:VEVENT
        ... UID:id1
        ... SUMMARY:Play Doom
        ... DTSTART:20050228T000000
        ... DURATION:PT2H
        ... END:VEVENT
        ... BEGIN:VEVENT
        ... UID:id3
        ... SUMMARY:Sleep (zzz)
        ... DTSTART:20050302T000000
        ... DURATION:PT8H
        ... END:VEVENT
        ... BEGIN:VEVENT
        ... UID:id4
        ... SUMMARY:Wake up
        ... DTSTART:20050302T080000
        ... DURATION:PT0H
        ... END:VEVENT
        ... END:VCALENDAR
        ... '''))

    The calendar now contains three events -- event2 is gone, event4 has
    appeared.

        >>> ids = [e.unique_id for e in cal]
        >>> ids.sort()
        >>> ids
        [u'id1', u'id3', u'id4']

    event1 is unchanged, and, in fact, it is the same object

        >>> cal.find('id1') is event1
        True

    event3 was modified in place

        >>> cal.find('id3') is event3
        True
        >>> event3.duration == timedelta(hours=8)
        True

    """


def test_suite():
    return unittest.TestSuite([
        doctest.DocTestSuite(setUp=setup.placelessSetUp,
                             tearDown=setup.placelessTearDown,
                             optionflags=doctest.ELLIPSIS),
        doctest.DocTestSuite('schooltool.app.cal',
                             optionflags=doctest.ELLIPSIS),
           ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Lyceum person specific code.
"""
from zope.catalog.text import TextIndex
from zope.index.text.interfaces import ISearchableText
from zope.interface import implements
from zope.component import adapts

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.person.person import Person
from schooltool.person.interfaces import IPersonFactory
from schooltool.course.section import PersonInstructorsCrowd
from schooltool.person.person import PersonCalendarCrowd
from schooltool.table.catalog import IndexedLocaleAwareGetterColumn
from schooltool.table.table import url_cell_formatter
from schooltool.relationship import RelationshipProperty
from schooltool.basicperson.advisor import URIAdvisor, URIAdvising, URIStudent
from schooltool.basicperson.interfaces import IBasicPerson
from schooltool.app.catalog import AttributeCatalog
from schooltool.common import SchoolToolMessage as _


class BasicPerson(Person):
    implements(IBasicPerson)

    prefix = None
    middle_name = None
    suffix = None
    preferred_name = None
    gender = None
    birth_date = None

    def __init__(self, username, first_name, last_name):
        self.first_name = first_name
        self.last_name = last_name
        self.username = username

    @property
    def title(self):
        return "%s %s" % (self.first_name, self.last_name)

    advisors = RelationshipProperty(rel_type=URIAdvising,
                                    my_role=URIStudent,
                                    other_role=URIAdvisor)

    advisees = RelationshipProperty(rel_type=URIAdvising,
                                    my_role=URIAdvisor,
                                    other_role=URIStudent)


class PersonFactoryUtility(object):

    implements(IPersonFactory)

    @property
    def name_sorting(self):
        app = ISchoolToolApplication(None)
        preferences = IApplicationPreferences(app)
        return preferences.name_sorting

    @property
    def columns_order(self):
        result = ['last_name', 'first_name']
        if self.name_sorting == 'first_name':
            result = list(reversed(result))
        return result

    def columns(self):
        first_name = IndexedLocaleAwareGetterColumn(
            index='first_name',
            name='first_name',
            cell_formatter=url_cell_formatter,
            title=_(u'First Name'),
            getter=lambda i, f: i.first_name,
            subsort=True)
        last_name = IndexedLocaleAwareGetterColumn(
            index='last_name',
            name='last_name',
            cell_formatter=url_cell_formatter,
            title=_(u'Last Name'),
            getter=lambda i, f: i.last_name,
            subsort=True)
        result = []
        for column_name in self.columns_order:
            result.append(locals()[column_name])
        return result

    def createManagerUser(self, username):
        return self(username, "Default", "Manager")

    def sortOn(self):
        result = []
        for column_name in self.columns_order:
            result.append((column_name, False))
        return tuple(result)

    def groupBy(self):
        return (("grade", False),)

    def __call__(self, *args, **kw):
        result = BasicPerson(*args, **kw)
        return result

    def getSortingKey(self, person, collator):
        result = []
        for column_name in self.columns_order:
            attr = getattr(person, column_name)
            result.append(collator.key(attr))
        return tuple(result)


class BasicPersonCalendarCrowd(PersonCalendarCrowd):
    """Crowd that allows instructor of a person access persons calendar.

    XXX write functional test.
    """
    adapts(IBasicPerson)

    def contains(self, principal):
        return (PersonCalendarCrowd.contains(self, principal) or
                PersonInstructorsCrowd(self.context).contains(principal))


class PersonCatalog(AttributeCatalog):

    version = '3 - updated title'
    interface = IBasicPerson
    attributes = ('__name__', 'title', 'first_name', 'last_name')

    def setIndexes(self, catalog):
        super(PersonCatalog, self).setIndexes(catalog)
        catalog['text'] = TextIndex('getSearchableText', ISearchableText, True)


getPersonCatalog = PersonCatalog.get


class SearchableTextPerson(object):

    adapts(IBasicPerson)
    implements(ISearchableText)

    def __init__(self, context):
        self.context = context

    def getSearchableText(self):
        result = [self.context.first_name, self.context.last_name,
                  self.context.username]
        return ' '.join(result)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Functional selenium tests for schooltool.group
"""
import unittest

from schooltool.testing.selenium import collect_ftests
from schooltool.group.stesting import group_selenium_layer


def test_suite():
    return collect_ftests(layer=group_selenium_layer)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Person interfaces
"""
import calendar
import pytz

from zope.container.interfaces import IContained, IContainer
from zope.container.constraints import containers, contains
from zope.annotation.interfaces import IAnnotatable, IAttributeAnnotatable
from zope.file.interfaces import IFile
from zope.interface import Interface, Attribute
from zope.schema import Bool, Choice, Object, TextLine

from schooltool.app.utils import vocabulary
from schooltool.common import SchoolToolMessage as _
from schooltool.group.interfaces import IGroupMember


class IPasswordWriter(Interface):
    """Interface for setting a password for a person."""

    def setPassword(password):
        """Set the password of the person."""


class IHavePreferences(IAnnotatable):
    """An object that can have preferences. Namely a Person."""


class IReadPerson(IGroupMember):
    """Publically accessible part of IPerson."""

    title = TextLine(
        title=_("Full name"),
        description=_("Name that should be displayed"))

    photo = Object(
        title=_("Photo"),
        required=False,
        schema=IFile)

    username = TextLine(
        title=_("Username"))

    # XXX: Should not be here. At most this could be in an adapter.
    overlaid_calendars = Attribute(
        """Additional calendars to overlay.

            A user may select a number of calendars that should be displayed in
            the calendar view, in addition to the user's calendar.

            This is a relationship property.""") # XXX Please use a Field.

    def checkPassword(password):
        """Check if the provided password is the same as the one set
        earlier using setPassword.

        Returns True if the passwords match.
        """

    def hasPassword():
        """Check if the user has a password.

        You can remove user's password (effectively disabling the account) by
        passing None to setPassword.  You can reenable the account by passing
        a new password to setPassword.
        """

    def __eq__(other):
        """Compare two persons."""


class IWritePerson(Interface):
    """Protected part of IPerson."""

    def setPassword(password):
        """Set the password in a hashed form, so it can be verified later.

        Setting password to None disables the user account.
        """


class IPerson(IReadPerson, IWritePerson, IAttributeAnnotatable):
    """Person.

    A person has a number of informative fields such as name, an optional
    photo, and so on.

    A person can also be a user of the system, therefore IPerson defines
    `username`, and methods for setting/checking passwords (`setPassword`,
    `checkPassword` and `hasPassword`).

    Use IPersonContained instead if you want a person in context (that is,
    one that you can use to traverse to other persons/groups/resources).
    """


class IPersonFactory(Interface):

    def __call__(*args, **kw):
        """Create a new Person instance."""

    def createManagerUser(username):
        """Create a person that will be the manager user.

        As different persons have different required fields and
        schooltool has to create the user automatically each person
        factory should have person specific manager user.
        """

    def columns():
        """Return a list of default columns for person lists."""

    def sortOn():
        """Return the default sort order for persons."""


class IPersonContainer(IContainer):
    """Container of persons."""

    contains(IPerson)

    super_user = Attribute(
        """Absolute administrator for this schooltool instance.

           A user that no matter which groups he is in or is not in has the
           administrative privileges.""")


class IPersonContained(IPerson, IContained):
    """Person contained in an IPersonContainer."""

    containers(IPersonContainer)


class ICalendarDisplayPreferences(Interface):
    """Preferences for displaying calendar events."""

    timezone = Choice(
        title=_("Time Zone"),
        values=pytz.common_timezones)

    timeformat = Choice(
        title=_("Time Format"),
        vocabulary=vocabulary([("%H:%M", _("HH:MM")),
                               ("%I:%M %p", _("HH:MM am/pm"))]))

    dateformat = Choice(
        title=_("Date Format"),
        vocabulary=vocabulary([("%m/%d/%y", _("MM/DD/YY")),
                               ("%Y-%m-%d", _("YYYY-MM-DD")),
                               ("%d %B, %Y", _("Day Month, Year"))]))

    # SUNDAY and MONDAY are integers, 6 and 0 respectivley
    weekstart = Choice(
        title=_("Week starts on"),
        vocabulary=vocabulary([(calendar.SATURDAY, _("Saturday")),
                               (calendar.SUNDAY, _("Sunday")),
                               (calendar.MONDAY, _("Monday"))]))


class IPersonPreferences(Interface):
    """Preferences stored in an annotation on a person."""

    __parent__ = Attribute(
        """Person who owns these preferences""")

    # XXX: Only available in SchoolTool, but that's ok for now.
    cal_periods = Bool(
        title=_("Daily calendar rows..."))

    cal_public = Bool(
        title=_("$PERSON calendar is visible to..."))



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.generations.evolve31
"""
import unittest
import doctest

from zope.app.testing import setup
from zope.interface import implements
from zope.container import btree
from zope.component import provideHandler

from schooltool.relationship.tests import setUpRelationships
from schooltool.generations.tests import ContextStub
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.basicperson.person import BasicPerson
from schooltool.relationship.tests import URIStub
from schooltool.relationship.interfaces import IRelationshipLinks
from schooltool.relationship.interfaces import IRelationshipRemovedEvent
from schooltool.relationship.interfaces import IRelationshipAddedEvent
from schooltool.relationship.relationship import relate
from schooltool.contact.contact import URIContact, URIPerson


class AppStub(btree.BTreeContainer):
    implements(ISchoolToolApplication)


def printRelationshipAdded(event):
    print 'Relate (%s): %s with %s' % (
        event.rel_type, event[URIPerson].first_name, event[URIContact].first_name)


def printRelationshipRemoved(event):
    print 'Unrelate (%s): %s from %s' % (
        event.rel_type, event[URIPerson].first_name, event[URIContact].first_name)


def doctest_evolve31():
    """Evolution to generation 31.

        >>> context = ContextStub()
        >>> context.root_folder['app'] = app = AppStub()

    Create a bunch of persons.

        >>> persons = app['persons'] = btree.BTreeContainer()
        >>> persons['will'] = BasicPerson("will", "William", "")
        >>> persons['vlad'] = BasicPerson("vlad", "Vladimir", "")
        >>> persons['john'] = BasicPerson("john", "Johny", "")
        >>> persons['pete'] = BasicPerson("pete", "Petey", "")
        >>> persons['bill'] = BasicPerson("bill", "Billy", "")

    Set up contacts.  In reality, contacts are instances of
    schooltool.contact.contact.Contact,  but we'll use BasicPerson as a stub.

        >>> def contact(person, contact, rel_type=URIContact):
        ...     relate(rel_type,
        ...            (person, URIPerson),
        ...            (contact, URIContact))

        >>> contact(persons['john'], persons['pete'])
        >>> contact(persons['john'], persons['bill'])
        >>> contact(persons['will'], persons['bill'])

    Add relationships that will not be evolved.

        >>> URIBoring = URIStub('Uninteresting relationship')
        >>> contact(persons['will'], persons['pete'], rel_type=URIBoring)

    Let's evolve now.

        >>> provideHandler(printRelationshipAdded,
        ...                [IRelationshipAddedEvent])
        >>> provideHandler(printRelationshipRemoved,
        ...                [IRelationshipRemovedEvent])

        >>> from schooltool.generations.evolve31 import evolve
        >>> evolve(context)
        Unrelate (<URIObject Contact>): Johny from Petey
        Relate (<URIObject Contact relationship>): Johny with Petey
        Unrelate (<URIObject Contact>): Johny from Billy
        Relate (<URIObject Contact relationship>): Johny with Billy
        Unrelate (<URIObject Contact>): William from Billy
        Relate (<URIObject Contact relationship>): William with Billy

    New relationships now have extra_info assigned.

        >>> links = IRelationshipLinks(persons['john']).getLinksByRole(URIContact)
        >>> for link in links:
        ...     print '%s %s: %s' % (
        ...         link.target.first_name, link.rel_type, link.extra_info)
        Petey <URIObject Contact relationship>:
          <schooltool.contact.contact.ContactPersonInfo object at ...>
        Billy <URIObject Contact relationship>:
          <schooltool.contact.contact.ContactPersonInfo object at ...>

    """


def setUp(test):
    setup.placelessSetUp()
    setup.setUpAnnotations()
    setUpRelationships()


def tearDown(test):
    setup.placelessTearDown()


def test_suite():
    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_ONLY_FIRST_FAILURE)
    return doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                optionflags=optionflags)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.securitypolicy.metaconfigure
"""
import unittest
import doctest

from zope.interface import implements, Interface
from zope.interface.verify import verifyObject
from zope.app.testing import setup
from zope.component import getAdapter, provideUtility
from zope.component import queryUtility

from schooltool.securitypolicy import metaconfigure as mc
from schooltool.securitypolicy.crowds import CrowdsUtility
from schooltool.securitypolicy.interfaces import ICrowdsUtility


def doctest_CrowdsUtility():
    """Doctest for CrowdsUtility.

        >>> cru = CrowdsUtility()
        >>> verifyObject(ICrowdsUtility, cru)
        True

    """


def doctest_getCrowdsUtility():
    """Doctest for getCrowdsUtility.

        >>> from schooltool.securitypolicy.metaconfigure import getCrowdsUtility
        >>> queryUtility(ICrowdsUtility) is None
        True

        >>> cru = getCrowdsUtility()
        >>> print cru
        <schooltool.securitypolicy.crowds.CrowdsUtility object ...>

        >>> queryUtility(ICrowdsUtility) is cru
        True

    """


def doctest_registerCrowdAdapter():
    """Doctests for registerCrowdAdapter.

        >>> from schooltool.securitypolicy import metaconfigure
        >>> from schooltool.securitypolicy.interfaces import ICrowd
        >>> cru = CrowdsUtility()

        >>> cru.factories = {
        ...     'crowd A': 'Factory A',
        ...     'crowd B': 'Factory B',
        ...     }

        >>> provideUtility(cru, provides=ICrowdsUtility)

        >>> class IMyObject(Interface):
        ...     pass

    Let's invoke registerCrowdAdapter:

        >>> metaconfigure.registerCrowdAdapter('perm', IMyObject)

    An adapter should have been registered.

        >>> class MyObject(object):
        ...     implements(IMyObject)
        >>> obj = MyObject()
        >>> adapter = getAdapter(obj, ICrowd, name='perm')
        >>> adapter
        <AggregateCrowd crowds=[]>

        >>> cru.crowds = {('perm', IMyObject): ['crowd A', 'crowd B']}
        >>> adapter.crowdFactories()
        ['Factory A', 'Factory B']

    """


def doctest_handle_crowd():
    """Doctest for handle_crowd

        >>> from schooltool.securitypolicy.metaconfigure import handle_crowd
        >>> cru = CrowdsUtility()
        >>> provideUtility(cru, provides=ICrowdsUtility)

        >>> handle_crowd('cr1', 'fac1')
        >>> print sorted(cru.factories.keys())
        ['cr1']
        >>> cru.factories['cr1']
        'fac1'
    """


def doctest_handle_allow():
    """Tests for handle_allow.

        >>> cru = CrowdsUtility()
        >>> cru.factories = {'cr1': 'Factory 1',
        ...                  'cr2': 'Factory 2',
        ...                  'cr3': 'Factory 3'}

        >>> provideUtility(cru, provides=ICrowdsUtility)
        >>> def registerCrowdAdapterStub(permission, iface):
        ...     print 'registerCrowdAdapter' + str((iface, permission))

        >>> mc.registerCrowdAdapter = registerCrowdAdapterStub

    First check a simple declaration:

        >>> mc.handle_allow('cr1', 'my.permission', 'iface')
        registerCrowdAdapter('iface', 'my.permission')
        >>> mc.handle_allow('cr2', 'my.permission', 'iface')

        >>> def printCrowds():
        ...     for discriminator, names in sorted(cru.crowds.items()):
        ...         print '%s: %s' % (discriminator, names)

        >>> printCrowds()
        ('my.permission', 'iface'): ['cr1', 'cr2']

    Another call will not invoke registerCrowdAdapter again:

        >>> mc.handle_allow('cr3', 'my.permission', 'iface')
        >>> printCrowds()
        ('my.permission', 'iface'): ['cr1', 'cr2', 'cr3']

    Crowd factories for the registered permission/interface pair
    can be obtained:

        >>> cru.getFactories('my.permission', 'iface')
        ['Factory 1', 'Factory 2', 'Factory 3']

    Passing a name that does not exist yet to the allow declaration
    will work as the actual lookup is postponed:

        >>> mc.handle_allow('cr4', 'my.permission', 'iface')
        >>> printCrowds()
        ('my.permission', 'iface'): ['cr1', 'cr2', 'cr3', 'cr4']

    But if you try to look up crowd factories you will get a key error:

        >>> cru.getFactories('my.permission', 'iface')
        Traceback (most recent call last):
        ...
        CrowdNotRegistered: cr4

    Let's check the case when an interface is not provided:

        >>> mc.handle_allow('cr1', 'my.permission', None)
        >>> printCrowds()
        ('my.permission', None): ['cr1']
        ('my.permission', 'iface'): ['cr1', 'cr2', 'cr3', 'cr4']

        >>> mc.handle_allow('cr2', 'my.permission', None)
        >>> printCrowds()
        ('my.permission', None): ['cr1', 'cr2']
        ('my.permission', 'iface'): ['cr1', 'cr2', 'cr3', 'cr4']

        >>> cru.getFactories('my.permission', None)
        ['Factory 1', 'Factory 2']

    """


class ContextStub(object):
    def action(self, discriminator, callable, args):
        print discriminator
        print callable
        print args
        print '---'


def doctest_crowd():
    """Doctests for crowd

        >>> from schooltool.securitypolicy.metaconfigure import crowd

        >>> _context = ContextStub()
        >>> crowd(_context, 'ipqs', 'ipqsfactory')
        ('crowd', 'ipqs')
          <function handle_crowd at ...>
          ('ipqs', 'ipqsfactory')
        ---

    """


def doctest_handle_aggregate_crowd():
    """Doctests for handle_aggregate_crowd.

        >>> cru = CrowdsUtility()
        >>> cru.factories['old1'] = 'crowd A'
        >>> cru.factories['old2'] = 'crowd B'
        >>> provideUtility(cru, provides=ICrowdsUtility)

        >>> from schooltool.securitypolicy import metaconfigure
        >>> metaconfigure.handle_aggregate_crowd('newcrowd', ['old1', 'old2'])

        >>> factory = cru.factories['newcrowd']
        >>> factory(None).crowdFactories()
        ['crowd A', 'crowd B']

    """


def doctest_aggregate_crowd():
    """Doctests for aggregate_crowd.

        >>> from schooltool.securitypolicy.metaconfigure import aggregate_crowd
        >>> _context = ContextStub()
        >>> aggregate_crowd(_context, 'newcrowd', ['old1', 'old2'])
        ('crowd', 'newcrowd')
          <function handle_aggregate_crowd ...>
          ('newcrowd', ['old1', 'old2'])
        ---

    """


def doctest_allow():
    """Doctests for allow.

        >>> from schooltool.securitypolicy.metaconfigure import allow
        >>> _context = ContextStub()
        >>> allow(_context, 'interface', ['ipqs', 'ecug'], 'do')
        ('allow', 'ipqs', 'do', 'interface')
          <function handle_allow ...>
          ('ipqs', 'do', 'interface')
        ---
        ('allow', 'ecug', 'do', 'interface')
          <function handle_allow ...>
          ('ecug', 'do', 'interface')
        ---

    """


def doctest_AccessControlSetting():
    """Doctests for AccessControlSetting.

        >>> from schooltool.securitypolicy.metaconfigure import AccessControlSetting
        >>> setting = AccessControlSetting("key", "Some text", "Alt text", False)

    Setting should implement the IAccessControlSetting interface:

        >>> from schooltool.securitypolicy.interfaces import IAccessControlSetting
        >>> verifyObject(IAccessControlSetting, setting)
        True

    getValue is just a wrapper for the get method of the
    adapter from schooltool application to IAccessControlCustomisations:

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> from schooltool.securitypolicy.interfaces import IAccessControlCustomisations
        >>> class CustomisationsStub(object):
        ...     implements(IAccessControlCustomisations)
        ...     def get(self, key):
        ...         return "Value for setting: %s" % key
        ...     def set(self, key, value):
        ...         return "Setting value of %s to %s" % (key, value)

        >>> class AppStub(object):
        ...     implements(ISchoolToolApplication)
        ...     def __conform__(self, iface):
        ...         if iface == IAccessControlCustomisations:
        ...             return CustomisationsStub()

        >>> from zope.component import provideAdapter
        >>> provideAdapter(lambda context: AppStub(),
        ...                adapts=[None],
        ...                provides=ISchoolToolApplication)

        >>> setting.getValue()
        'Value for setting: key'

    We can set the value of the setting directly as well:

        >>> setting.setValue("new value")
        'Setting value of key to new value'

    """


def doctest_handle_setting():
    """Tests for handle_setting.

        >>> from schooltool.securitypolicy.metaconfigure import handle_setting

    Handle setting should register a subscriber adapter for
    IAccessControlSetting interface, initially there are no subscribers:

        >>> from zope.component import subscribers
        >>> from schooltool.securitypolicy.interfaces import IAccessControlSetting
        >>> subscribers([None], IAccessControlSetting)
        []

        >>> handle_setting("key", "text", None, False)

    Now we should have one:

        >>> subscribers([None], IAccessControlSetting)
        [<AccessControlSetting key=key, text=text, default=False>]

    Let's add another one:

        >>> handle_setting("another_key", "more text", "alt text", True)
        >>> subscribers([None], IAccessControlSetting)
        [<AccessControlSetting key=key, text=text, default=False>,
         <AccessControlSetting key=another_key, text=more text,
                               alt_text=alt text, default=True>]

    """


class Fixture(object):

    def setUp(self, test=None):
        setup.placelessSetUp()
        self.oldRegisterCrowdAdapter = mc.registerCrowdAdapter

    def tearDown(self, test=None):
        setup.placelessTearDown()
        mc.registerCrowdAdapter = self.oldRegisterCrowdAdapter


def test_suite():
    f = Fixture()
    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_NDIFF)
    return unittest.TestSuite([
            doctest.DocTestSuite(optionflags=optionflags,
                                 setUp=f.setUp, tearDown=f.tearDown)])




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
School year implementation
"""
import datetime

from zope.proxy import sameProxiedObjects
from zope.event import notify
from zope.component import queryUtility
from zope.component import adapter
from zope.component import adapts
from zope.interface import implementer
from zope.interface import implements
from zope.container.btree import BTreeContainer

from schooltool.common import DateRange
from schooltool.common import IDateRange
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.app import InitBase
from schooltool.term.interfaces import IDateManager
from schooltool.term.interfaces import ITermContainer
from schooltool.schoolyear.subscriber import EventAdapterSubscriber
from schooltool.schoolyear.interfaces import TermOverflowError
from schooltool.schoolyear.interfaces import TermOverlapError
from schooltool.schoolyear.interfaces import SchoolYearOverlapError
from schooltool.schoolyear.interfaces import ISubscriber
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.schoolyear.interfaces import ISchoolYearContainer

SCHOOLYEAR_CONTAINER_KEY = 'schooltool.schoolyear'


class SchoolYearBeforeChangeEvent(object):

    def __init__(self, schoolyear, old_dates, new_dates):
        self.schoolyear = schoolyear
        self.old_dates = old_dates
        self.new_dates = new_dates


class SchoolYearAfterChangeEvent(object):

    def __init__(self, schoolyear, old_dates, new_dates):
        self.schoolyear = schoolyear
        self.old_dates = old_dates
        self.new_dates = new_dates


class SchoolYearContainer(BTreeContainer):
    implements(ISchoolYearContainer)

    _active_id = None

    def _set_active_id(self, new_id):
        if new_id is not None and new_id not in self:
            raise ValueError("School Year %r does not exist" % new_id)
        self._active_id = new_id

    active_id = property(lambda self: self._active_id)

    def __delitem__(self, schoolyear_id):
        if schoolyear_id == self.active_id:
            if len(self.values()) > 1:
                raise ValueError("Can not delete an active schoolyear, unless"
                                 " it is the last school year available!")
            else:
                self._set_active_id(None)
        BTreeContainer.__delitem__(self, schoolyear_id)

    def validateForOverlap(self, schoolyear):
        overlapping_schoolyears = []
        for other_schoolyear in self.values():
            if IDateRange(schoolyear).overlaps(IDateRange(other_schoolyear)):
                overlapping_schoolyears.append(other_schoolyear)

        if overlapping_schoolyears:
            raise SchoolYearOverlapError(schoolyear, overlapping_schoolyears)

    def __setitem__(self, key, schoolyear):
        self.validateForOverlap(schoolyear)
        BTreeContainer.__setitem__(self, key, schoolyear)
        if self.active_id is None:
            self._set_active_id(key)

    def getActiveSchoolYear(self):
        if self.active_id:
            return self[self.active_id]
        return None

    @property
    def sorted_schoolyears(self):
        return sorted(self.values(), key=lambda s: s.last)

    def activateNextSchoolYear(self, year_id=None):
        if year_id is None:
            next = self.getNextSchoolYear()
            if next is not None:
                year_id = next.__name__
        self._set_active_id(year_id)

    def getLastSchoolYearForDate(self, date):
        before, after = [], []
        for year in self.sorted_schoolyears:
            if date in IDateRange(year):
                return year
            if date > year.last:
                before.append((year.last, year))
            if date < year.first:
                after.append((year.first, year))
        if before:
            return max(before)[1]
        if after:
            return min(after)[1]
        return None

    def getSchoolYearForToday(self):
        dtm = queryUtility(IDateManager)
        today = dtm.today
        for year in self.sorted_schoolyears:
            if today in IDateRange(year):
                return year
        return None

    def getNextSchoolYear(self):
        if self.getActiveSchoolYear() is None:
            return None

        years = self.sorted_schoolyears
        this_schoolyear_index = years.index(self.getActiveSchoolYear())
        next_schoolyear_index = this_schoolyear_index + 1
        if next_schoolyear_index < len(years):
            return years[next_schoolyear_index]

        return None


class SchoolYear(BTreeContainer):
    implements(ITermContainer, ISchoolYear)

    def __init__(self, title, first, last):
        self.title = title
        self._first = first
        self._last = last
        if last < first:
            raise ValueError("Last date %r less than first date %r" %
                             (last, first))
        BTreeContainer.__init__(self)

    @property
    def first(self):
        return self._first

    @first.setter
    def first(self, new_first_date):
        old_dates = (self._first, self._last)
        new_dates = (new_first_date, self._last)

        if self._last < new_first_date:
            raise ValueError("Last date %r less than first date %r" %
                             (self._last, new_first_date))

        notify(SchoolYearBeforeChangeEvent(self, old_dates, new_dates))
        self._first = new_first_date
        notify(SchoolYearAfterChangeEvent(self, old_dates, new_dates))

    @property
    def last(self):
        return self._last

    @last.setter
    def last(self, new_last_date):
        old_dates = (self._first, self._last)
        new_dates = (self._first, new_last_date)

        if new_last_date < self._first:
            raise ValueError("Last date %r less than first date %r" %
                             (new_last_date, self._first))

        notify(SchoolYearBeforeChangeEvent(self, old_dates, new_dates))
        self._last = new_last_date
        notify(SchoolYearAfterChangeEvent(self, old_dates, new_dates))

    def validateForOverlap(self, term):
        overlapping_terms = []
        for other_term in self.values():
            if term.overlaps(other_term):
                overlapping_terms.append(other_term)

        if overlapping_terms:
            raise TermOverlapError(term, overlapping_terms)

    def __setitem__(self, key, term):
        self.validateForOverlap(term)
        if term.first < self.first:
            raise ValueError("Term can't start before the school year starts!")
        if term.last > self.last:
            raise ValueError("Term can't end after the school year ends!")
        BTreeContainer.__setitem__(self, key, term)


class SchoolYearDateRangeAdapter(DateRange):
    adapts(ISchoolYear)
    implements(IDateRange)

    def __init__(self, context):
        self.context = context

    @property
    def first(self):
        return self.context.first

    @first.setter
    def first(self, new_first_date):
        self.context.first = new_first_date

    @property
    def last(self):
        return self.context.last

    @last.setter
    def last(self, new_last_date):
        self.context.first = new_last_date


@adapter(ISchoolToolApplication)
@implementer(ISchoolYearContainer)
def getSchoolYearContainer(app):
    return app[SCHOOLYEAR_CONTAINER_KEY]


class SchoolYearInit(InitBase):

    adapts(ISchoolToolApplication)

    def __call__(self):
        self.app[SCHOOLYEAR_CONTAINER_KEY] = SchoolYearContainer()


def validateScholYearsForOverlap(syc, dr, schoolyear):
    overlapping_schoolyears = []
    for other_schoolyear in syc.values():
        if (not sameProxiedObjects(other_schoolyear, schoolyear) and
            dr.overlaps(IDateRange(other_schoolyear))):
            overlapping_schoolyears.append(other_schoolyear)
            if overlapping_schoolyears:
                raise SchoolYearOverlapError(schoolyear,
                                             overlapping_schoolyears)


class SchoolYearOverlapValidationSubscriber(EventAdapterSubscriber):
    adapts(SchoolYearBeforeChangeEvent)
    implements(ISubscriber)

    def __call__(self):
        syc = self.event.schoolyear.__parent__
        dr = DateRange(*self.event.new_dates)
        if syc:
            validateScholYearsForOverlap(syc, dr, self.event.schoolyear)


def validateScholYearForOverflow(dr, schoolyear):
    overflowing_terms = []
    for term in schoolyear.values():
        if (term.first not in dr or
            term.last not in dr):
            overflowing_terms.append(term)
    if overflowing_terms:
        raise TermOverflowError(schoolyear, overflowing_terms)


class SchoolYearTermOverflowValidationSubscriber(EventAdapterSubscriber):
    adapts(SchoolYearBeforeChangeEvent)
    implements(ISubscriber)

    def __call__(self):
        overflowing_terms = []
        sy = self.event.schoolyear
        dr = DateRange(*self.event.new_dates)
        validateScholYearForOverflow(dr, sy)


@adapter(datetime.date)
@implementer(ITermContainer)
def getTermContainerForDate(date):
    app = ISchoolToolApplication(None)
    syc = ISchoolYearContainer(app)
    year = syc.getLastSchoolYearForDate(date)
    term_container = ITermContainer(year, None)
    return term_container





#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Course CSV import
"""
from decimal import Decimal, InvalidOperation

from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL

import schooltool.skin.flourish.page
from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.browser.csvimport import BaseCSVImporter
from schooltool.app.browser.csvimport import BaseCSVImportView
from schooltool.app.browser.csvimport import FlourishBaseCSVImportView
from schooltool.course.course import Course
from schooltool.course.interfaces import ICourseContainer
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.skin import flourish

from schooltool.common import SchoolToolMessage as _


class CourseCSVImporter(BaseCSVImporter):
    """Course CSV Importer"""

    factory = Course

    def createAndAdd(self, data, dry_run=True):
        """Create objects and add them to the container."""

        course_id = None
        government_id = None
        credits = None

        if len(data) < 1:
            self.errors.fields.append(_('Insufficient data provided.'))
            return

        if not data[0]:
            self.errors.fields.append(_('Titles may not be empty'))
            return

        if len(data) > 1:
            description = data[1]
        else:
            description = ''

        if len(data) > 2:
            if data[2]:
                course_id = data[2]
        else:
            for key, course in self.container.items():
                if course.title == data[0]:
                    course_id = key
                    break

        if course_id in self.container:
            obj = self.container[course_id]
            name = course_id
        else:
            obj = None

        if len(data) > 3:
            if data[3]:
                government_id = data[3]

        if len(data) > 4:
            try:
                credits = Decimal(data[4])
            except (ValueError, InvalidOperation):
                self.errors.fields.append(_('Course "${course_title}" credits "${invalid_credits}" value'
                                            ' must be a number.',
                                            mapping={'course_title': data[0],
                                                     'invalid_credits': data[4]}))
                return

        if not obj:
            obj = self.factory(title=data[0], description=description)
            obj.course_id = course_id
            obj.government_id = government_id
            obj.credits = credits
            try:
                name = self.chooser.chooseName(course_id, obj)
            except ValueError, e:
                msg = e.args[0]
                self.errors.fields.append(_('Course "${course_title}" id "${invalid_id}"'
                                            ' is invalid. ${error_message}',
                                            mapping={'course_title': obj.title,
                                                     'invalid_id': course_id,
                                                     'error_message': msg}))
                return

        if not dry_run:
            if course_id in self.container:
                obj.title = data[0]
                obj.description = description
                obj.course_id = course_id
                obj.government_id = government_id
                obj.credits = credits
            else:
                self.container[name] = obj


class CourseCSVImportView(BaseCSVImportView):
    """View for Course CSV importer."""

    importer_class = CourseCSVImporter


class FlourishCourseCSVImportView(FlourishBaseCSVImportView, ActiveSchoolYearContentMixin):

    importer_class = CourseCSVImporter

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    def nextURL(self):
        app = ISchoolToolApplication(None)
        return self.url_with_schoolyear_id(app, view_name='courses')


class ImportCoursesLinkViewlet(flourish.page.LinkViewlet,
                               ActiveSchoolYearContentMixin):

    @property
    def enabled(self):
        courses = ICourseContainer(self.schoolyear)
        if not flourish.canEdit(courses):
            return False
        return super(ImportCoursesLinkViewlet, self).enabled

    @property
    def url(self):
        link = self.link
        if not link:
            return None
        courses = ICourseContainer(self.schoolyear)
        return "%s/%s" % (absoluteURL(courses, self.request),
                          self.link)


class SectionMemberCSVImporter(BaseCSVImporter):
    """Section Member CSV Importer"""

    def createAndAdd(self, data, dry_run=True):
        """Create persons and add them to the section learners."""

        if len(data) < 1:
            self.errors.fields.append(_('Insufficient data provided.'))
            return

        if not data[0]:
            self.errors.fields.append(_('User names must not be empty.'))
            return

        app = ISchoolToolApplication(None)
        person_container = app['persons']
        username = data[0]

        if username not in person_container:
            self.errors.fields.append(_('"${username}" is not a valid username.',
                                        mapping={'username': username}))
            return

        user = person_container[username]
        if not dry_run:
            removeSecurityProxy(self.container.members).add(removeSecurityProxy(user))


class SectionMemberCSVImportView(BaseCSVImportView):
    """View for Section Member CSV importer."""

    importer_class = SectionMemberCSVImporter


class FlourishSectionMemberCSVImportView(FlourishBaseCSVImportView):

    importer_class = SectionMemberCSVImporter

    def nextURL(self):
        return absoluteURL(self.context, self.request)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool security policy testing helpers.
"""

import zope.schema

from schooltool.securitypolicy.metaconfigure import getDescriptionUtility


def discriminator_sort_key(disc):
    if disc[1] is None:
        return ('', 'None', disc[0])
    return (str(disc[1].__module__), str(disc[1].__name__), disc[0])


def collectActionsByDiscriminator():
    util = getDescriptionUtility()
    collected = {}
    for group in util.actions_by_group.values():
        for action in group.values():
            discriminator = (action.permission, action.interface)
            if discriminator not in collected:
                collected[discriminator] = []
            collected[discriminator].append(action)
    for actions in collected.values():
        actions[:] = sorted(actions,
                            key=lambda a: a.__name__ + a.__parent__.__name__)
    return collected


def printActionDescriptions(actions_by_discriminator):
    last_module = ''
    util = getDescriptionUtility()
    for disc in sorted(actions_by_discriminator, key=discriminator_sort_key):
        mod, ifc, perm = discriminator_sort_key(disc)
        if mod != last_module:
            last_module = mod
            print '=' * len(last_module)
            print last_module
            print '=' * len(last_module)
        perm_pair_desc = '%s, %s' % (ifc, perm)
        print '- %s\n- %s' % (perm_pair_desc, '-' * len(perm_pair_desc))
        listed = [
            str('%s / %s' % (util.groups[a.__parent__.__name__].title,
                             a.title))
            for a in actions_by_discriminator[disc]]
        for act in listed:
            print '-  %s' % act
        print '-'


def printDiscriminators(discriminators):
    last_module = ''
    for disc in sorted(discriminators, key=discriminator_sort_key):
        mod, ifc, perm = discriminator_sort_key(disc)
        if mod != last_module:
            last_module = mod
            print '-' * len(last_module)
            print last_module
            print '-' * len(last_module)
        print '%s, %s' % (ifc, perm)


def printDirectiveDescription(interface):
    for name, field in zope.schema.getFieldsInOrder(interface):
        title = '%s %s' % (name,
                           field.required and '(required)' or '(optional)')
        spacing = ' ' * (len(title) + 2)
        desc = ('\n' + spacing).join(
            [ln.strip() for ln in field.description.strip().splitlines()])
        print title + (desc and ': ' or '') + desc



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Generations for database version upgrades.
"""

from zope.app.generations.generations import SchemaManager

schemaManager = SchemaManager(
    minimum_generation=42,
    generation=42,
    package_name='schooltool.generations')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys, imp

from zope.testing import cleanup


class ModulesSnapshot(object):

    def __init__(self):
        self.originals = dict(sys.modules)
        self.faked_attrs = {}
        self.imported = set()

    def mock(self, values):
        """Usage:
        mock({'my.module.Foo': FooClass,
              'my.module.another.GLOBAL_VAL': 1,
              })
        """
        for full_name in sorted(values):
            names = full_name.split('.')
            attr_name = names[-1]
            module_name = '.'.join(names[:-1])
            assert module_name
            self.mock_attr(module_name, attr_name, fake=values[full_name])

    def mock_attr(self, modulename, name, fake=None):
        module = self.get_module(modulename)
        if modulename not in self.faked_attrs:
            self.faked_attrs[modulename] = {
                'injected': [],
                'replaced': {},
                }
        faked = self.faked_attrs[modulename]
        if hasattr(module, name):
            if (name not in faked['injected'] and
                name not in faked['replaced']):
                faked['replaced'][name] = getattr(module, name)
        else:
            faked['injected'].append(name)
        setattr(module, name, fake)

    def mock_module(self, modulename, fake=None):
        if fake is None:
            fake = imp.new_module(modulename)
        names = modulename.split('.')
        this_name = names[-1]
        parent_name = '.'.join(names[:-1])
        if parent_name:
            self.mock_attr(parent_name, this_name, fake)
        sys.modules[modulename] = fake
        self.imported.add(modulename)

    def get_module(self, modulename):
        try:
            sys.modules[modulename]
        except:
            self.mock_module(modulename)
        return sys.modules[modulename]

    def restore(self):
        for module_name, faked in self.faked_attrs.items():
            try:
                module = sys.modules[module_name]
            except:
                continue
            for name in faked['injected']:
                del module.__dict__[name]
            for name, old_value in faked['replaced'].items():
                setattr(module, name, old_value)

        for name in list(sys.modules):
            if name in self.imported:
                del sys.modules[name]
            elif name in self.originals:
                sys.modules[name] = self.originals[name]


_snapshot = None

def getSnapshot():
    global _snapshot
    if _snapshot is None:
        _snapshot = ModulesSnapshot()
    return _snapshot

def restoreModules():
    global _snapshot
    if _snapshot is not None:
        _snapshot.restore()
        _snapshot = None


cleanup.addCleanUp(restoreModules)


def module(module_name):
    """A decorator to put method or class to the given module."""
    snapshot = getSnapshot()
    if type(module_name) == type(sys):
        module_name = module_name.__name__
    def mock_something(something):
        name = getattr(something, '__name__', None)
        snapshot.mock_attr(module_name, name, fake=something)
        return something
    return mock_something


def fake_global(module_name, name, value):
    """Set a global variable to a module."""
    snapshot = getSnapshot()
    if type(module_name) == type(sys):
        module_name = module_name.__name__
    snapshot.mock_attr(module_name, name, fake=value)


def fake_module(module_name, other=None):
    """Replace a module with a fake one."""
    snapshot = getSnapshot()
    if type(module_name) == type(sys):
        module_name = module_name.__name__
    snapshot.mock_module(module_name, fake=other)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
High-level setup functions for functional tests relating to basicperson..
"""

from schooltool.app.browser.ftests.setup import logInManager

def addPerson(firstname, lastname, username=None, password=None, groups=None, browser=None):
    """Add a person.

    If username is not specified, it will be taken to be "<firstname>.<lastname>".lower().

    If password is not specified, it will be taken to be username + 'pwd'.
    """
    if not username:
        username = ('%s.%s' % (firstname, lastname)).lower()
    if not password:
        password = username + 'pwd'
    if browser is None:
        browser = logInManager()
    browser.getLink('Manage').click()
    browser.getLink('Persons').click()
    browser.getLink('New Person').click()

    browser.getControl('First name').value = firstname
    browser.getControl('Last name').value = lastname
    browser.getControl('Username').value = username
    browser.getControl('Password').value = password
    browser.getControl('Confirm').value = password
    browser.getControl('Add').click()

    if groups:
        browser.open('http://localhost/persons/%s' % username)
        browser.getLink('edit groups').click()
        for group in groups:
            browser.getControl(name='add_item.%s' % group).value = True
        browser.getControl('Add').click()
    browser.open('http://localhost/persons')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from zope.component import queryUtility
from zope.traversing.api import traverse


class UtilitySpecification(object):
    """A specification of a utility.

    A utility gets:

    factory - the factory creating the utility.
    iface - the interface provided by the utility (by which it is looked up)
    name - optional utility name
    setUp - an optional setUp function for the utility. This can
            for instance be used for setting up indexes in a catalog.
            The setUp function gets passed the utility as an argument.
    override - override previous local utility setups. Replaces that
               utility with this one. Defaults to False
    storage_name - name the utility as visible in the ZMI.
                   If not specified will be autogenerated.
    """

    def __init__(self, factory, iface, name='', setUp=None,
                 override=False, storage_name=None):
        if storage_name is None:
            storage_name = iface.__module__ + '.' + iface.__name__
            if name:
                storage_name += '.' + name
        self.storage_name = storage_name
        self.factory = factory
        self.setUp = setUp
        self.iface = iface
        self.utility_name = name
        self.override = override


def setUpUtilities(site, specs):
   manager = site.getSiteManager()
   default = traverse(site, '++etc++site/default')
   for spec in specs:
       local_utility = getLocalUtility(default, spec)
       if local_utility is not None:
           if spec.override:
               # override existing utility
               name = local_utility.__name__
               manager.unregisterUtility(
                   local_utility, spec.iface, spec.utility_name)
               del default[name]
           else:
               # do not register this utility; we already got it
               continue
       utility = spec.factory()
       default[spec.storage_name] = utility
       if spec.setUp is not None:
           spec.setUp(utility)
       manager.registerUtility(utility, spec.iface, spec.utility_name)


def getLocalUtility(default, spec):
    util = queryUtility(spec.iface, name=spec.utility_name, default=None)
    if util is None or getattr(util, '__parent__', None) is default:
        return util
    else:
        return None


class UtilitySetUp(UtilitySpecification):
    """Set up a single utility."""

    def __call__(self, event):
        site = event.object
        setUpUtilities(site, [self])


class MultiUtilitySetUp(object):
    """Set up multiple related utilities that need to be in order.

    (for instance intids needs to be set up before the catalog is).
    """

    def __init__(self, *specifications):
        self.specifications = specifications

    def __call__(self, event):
        site = event.object
        setUpUtilities(site, self.specifications)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.generations.evolve29
"""
import unittest
import doctest

from zope.app.testing import setup
from zope.interface import implements

from schooltool.basicperson.person import BasicPerson
from schooltool.generations.tests import ContextStub
from schooltool.app.interfaces import ISchoolToolApplication


class AppStub(dict):
    implements(ISchoolToolApplication)

    def __init__(self):
        self['persons'] = {}
        self['persons']['john'] = BasicPerson("john", "Johny", "John")
        self['persons']['john'].email = "john@example.com"
        self['persons']['john'].phone = "667755"

        self['persons']['pete'] = BasicPerson("pete", "Petey", "Pete")
        self['persons']['pete'].email = "pete@example.com"
        self['persons']['pete'].phone = "667755"


def doctest_evolve29():
    """Evolution to generation 29.

        >>> context = ContextStub()
        >>> context.root_folder['app'] = app = AppStub()

    We set up some persons and evolve the application

        >>> from schooltool.generations.evolve29 import evolve
        >>> evolve(context)

    and the extra attributes are gone

        >>> for username, person in sorted(app['persons'].items()):
        ...     print person.__dict__
        {'username': 'john', 'first_name': 'Johny', 'last_name': 'John'}
        {'username': 'pete', 'first_name': 'Petey', 'last_name': 'Pete'}

    """


def setUp(test):
    setup.placelessSetUp()


def tearDown(test):
    setup.placelessTearDown()


def test_suite():
    return doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                optionflags=doctest.ELLIPSIS
                                |doctest.REPORT_ONLY_FIRST_FAILURE)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Registration of types of resources.
"""
from zope.component import adapts
from zope.component import queryUtility
from zope.interface import directlyProvides
from zope.interface import implements

from zc.table.column import GetterColumn
from zc.table.interfaces import ISortableColumn

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.common import SchoolToolMessage as _
from schooltool.resource.interfaces import (IResource, IResourceFactoryUtility,
                                            ILocation, IEquipment,
                                            IResourceTypeInformation,
                                            IResourceSubTypes)
from schooltool.table.table import LocaleAwareGetterColumn


class ResourceFactoryUtility(object):
    implements(IResourceSubTypes, IResourceFactoryUtility)

    def __init__(self):
        self.interface = IResource

    def types(self):
        app = ISchoolToolApplication(None)
        types = set()
        for resource in app['resources'].values():
            if self.interface.providedBy(resource):
                types.add(resource.type)
        return list(types)

    title = _("Resource")

    def columns(self):
        title = LocaleAwareGetterColumn(
            name='title',
            title=_(u'Title'),
            getter=lambda i, f: i.title,
            subsort=True)
        directlyProvides(title, ISortableColumn)
        return [title]

class LocationFactoryUtility(ResourceFactoryUtility):
    implements(IResourceFactoryUtility)

    title = _("Location")

    def __init__(self):
        self.interface = ILocation

    def columns(self):
        title = LocaleAwareGetterColumn(
            name='title',
            title=_(u'Title'),
            getter=lambda i, f: i.title,
            subsort=True)
        directlyProvides(title, ISortableColumn)
        capacity = GetterColumn(
            name='capacity',
            title=_(u'Capacity'),
            getter=lambda i, f: i.capacity,
            subsort=True)
        directlyProvides(capacity, ISortableColumn)
        return [title, capacity]


class EquipmentFactoryUtility(ResourceFactoryUtility):
    implements(IResourceFactoryUtility)

    title = _("Equipment")

    def __init__(self):
        self.interface = IEquipment

    def columns(self):
        title = LocaleAwareGetterColumn(
            name='title',
            title=_(u'Title'),
            getter=lambda i, f: i.title,
            subsort=True)
        directlyProvides(title, ISortableColumn)
        type = GetterColumn(
            name='type',
            title=_(u'Type'),
            getter=lambda i, f: i.type,
            subsort=True)
        directlyProvides(type, ISortableColumn)
        manufacturer = GetterColumn(
            name='manufacturer',
            title=_(u'Manufacturer'),
            getter=lambda i, f: i.manufacturer,
            subsort=True)
        directlyProvides(manufacturer, ISortableColumn)
        model = GetterColumn(
            name='model',
            title=_(u'Model'),
            getter=lambda i, f: i.model,
            subsort=True)
        directlyProvides(model, ISortableColumn)
        serialNumber = GetterColumn(
            name='serialNumber',
            title=_(u'Serial Number'),
            getter=lambda i, f: i.serialNumber,
            subsort=True)
        directlyProvides(serialNumber, ISortableColumn)
        purchaseDate = GetterColumn(
            name='purchaseDate',
            title=_(u'Purchase Date'),
            getter=lambda i, f: i.purchaseDate,
            subsort=True)
        directlyProvides(purchaseDate, ISortableColumn)
        return [title, type, manufacturer, model, serialNumber, purchaseDate]


class ResourceTypeAdapter(object):
    implements(IResourceTypeInformation)
    adapts(IResource)

    id = "resource"

    def __init__(self, context):
        self.context = context

    @property
    def title(self):
        return queryUtility(IResourceFactoryUtility, name=self.id).title


class LocationTypeAdapter(ResourceTypeAdapter):
    adapts(ILocation)

    id = "location"


class EquipmentTypeAdapter(ResourceTypeAdapter):
    adapts(IEquipment)

    id = "equipment"



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Section implementation
"""
from persistent import Persistent

from zope.annotation.interfaces import IAttributeAnnotatable
from zope.intid.interfaces import IIntIds
from zope.interface import implements
from zope.interface import implementer
from zope.event import notify
from zope.component import getUtility
from zope.component import adapter
from zope.component import adapts
from zope.container.interfaces import INameChooser
from zope.container.btree import BTreeContainer
from zope.container.contained import Contained
from zope.lifecycleevent.interfaces import IObjectRemovedEvent
from zope.proxy import sameProxiedObjects
from zope.security.proxy import removeSecurityProxy

from schooltool.app import membership
from schooltool.app.relationships import URIInstruction, URISection
from schooltool.app.app import InitBase
from schooltool.app import relationships
from schooltool.app.security import ConfigurableCrowd
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.utils import vocabulary_titled
from schooltool.course import interfaces, booking
from schooltool.group.interfaces import IBaseGroup as IGroup
from schooltool.person.interfaces import IPerson
from schooltool.relationship.relationship import getRelatedObjects
from schooltool.relationship import RelationshipProperty
from schooltool.schoolyear.subscriber import EventAdapterSubscriber
from schooltool.schoolyear.subscriber import ObjectEventAdapterSubscriber
from schooltool.schoolyear.interfaces import ISubscriber
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.securitypolicy.crowds import Crowd, AggregateCrowd
from schooltool.securitypolicy.crowds import ClerksCrowd, AdministratorsCrowd
from schooltool.term.term import getNextTerm
from schooltool.term.interfaces import ITerm

from schooltool.common import SchoolToolMessage as _


class InvalidSectionLinkException(Exception):
    pass


class SectionBeforeLinkingEvent(object):
    def __init__(self, first, second):
        self.first = first
        self.second = second


class Section(Persistent, Contained):

    implements(interfaces.ISectionContained,
               IAttributeAnnotatable)

    _location = None

    _previous = None
    _next = None

    def __init__(self, title="Section", description=None, schedule=None):
        self.title = title
        self.description = description

    @property
    def label(self):
        instructors = "; ".join([i.title for i in self.instructors])
        msg = _('${instructors} -- ${section_title}',
                mapping={'instructors': instructors, 'section_title': self.title})
        return msg

    @property
    def size(self):
        size = 0
        for member in self.members:
            if IPerson.providedBy(member):
                size = size + 1
            elif IGroup.providedBy(member):
                size = size + len(member.members)
        return size

    def _unlinkRangeTo(self, other):
        """Unlink sections between self and the other in self.linked_sections."""
        linked = self.linked_sections
        if other not in linked or self is other:
            return
        idx_first, idx_last = sorted([linked.index(self), linked.index(other)])
        linked[idx_first]._next = None
        for section in linked[idx_first+1:idx_last]:
            section._previous = None
            section._next = None
        linked[idx_last]._previous = None

    @property
    def previous(self):
        return self._previous

    @previous.setter
    def previous(self, new):
        new = removeSecurityProxy(new)
        if new is self._previous:
            return
        if new is self:
            raise InvalidSectionLinkException(
                _('Cannot assign section as previous to itself'))

        notify(SectionBeforeLinkingEvent(new, self))

        if new is not None:
            self._unlinkRangeTo(new)

        old_prev = self._previous
        self._previous = None
        if old_prev is not None:
            old_prev.next = None
        self._previous = new

        if new is not None:
            new.next = self

    @property
    def next(self):
        return self._next

    @next.setter
    def next(self, new):
        new = removeSecurityProxy(new)
        if new is self._next:
            return
        if new is self:
            raise InvalidSectionLinkException(
                _('Cannot assign section as next to itself'))

        notify(SectionBeforeLinkingEvent(self, new))

        if new is not None:
            self._unlinkRangeTo(new)

        old_next = self._next
        self._next = None
        if old_next is not None:
            old_next.previous = None
        self._next = new

        if new is not None:
            new.previous = self

    @property
    def linked_sections(self):
        sections = [self]

        pit = self.previous
        while pit:
            sections.insert(0, pit)
            pit = pit.previous

        nit = self.next
        while nit:
            sections.append(nit)
            nit = nit.next

        return sections


    instructors = RelationshipProperty(relationships.URIInstruction,
                                       relationships.URISection,
                                       relationships.URIInstructor)

    courses = RelationshipProperty(relationships.URICourseSections,
                                   relationships.URISectionOfCourse,
                                   relationships.URICourse)

    members = RelationshipProperty(membership.URIMembership,
                                   membership.URIGroup,
                                   membership.URIMember)

    resources = RelationshipProperty(booking.URISectionBooking,
                                     booking.URISection,
                                     booking.URIResource)


@adapter(ITerm)
@implementer(interfaces.ISectionContainer)
def getSectionContainer(term):
    int_ids = getUtility(IIntIds)
    term_id = str(int_ids.getId(term))
    app = ISchoolToolApplication(None)
    sc = app['schooltool.course.section'].get(term_id, None)
    if sc is None:
        sc = app['schooltool.course.section'][term_id] = SectionContainer()
    return sc


@adapter(interfaces.ISectionContainer)
@implementer(ISchoolYear)
def getSchoolYearForSectionContainer(section_container):
    return ISchoolYear(ITerm(section_container))


@adapter(interfaces.ISection)
@implementer(ISchoolYear)
def getSchoolYearForSection(section):
    return ISchoolYear(ITerm(section.__parent__))


@adapter(interfaces.ISectionContainer)
@implementer(ITerm)
def getTermForSectionContainer(section_container):
    container_id = int(section_container.__name__)
    int_ids = getUtility(IIntIds)
    container = int_ids.getObject(container_id)
    return container


@adapter(interfaces.ISection)
@implementer(ITerm)
def getTermForSection(section):
    return ITerm(section.__parent__)


@adapter(interfaces.ISectionContainer)
@implementer(interfaces.ICourseContainer)
def getCourseContainerForSectionContainer(section_container):
    return interfaces.ICourseContainer(ISchoolYear(section_container))


@adapter(interfaces.ISection)
@implementer(interfaces.ICourseContainer)
def getCourseContainerForSection(section):
    return interfaces.ICourseContainer(ISchoolYear(section))


class SectionContainerContainer(BTreeContainer):
    """Container of Section containers."""

    implements(interfaces.ISectionContainerContainer)


class SectionContainer(BTreeContainer):
    """Container of Sections."""

    implements(interfaces.ISectionContainer, IAttributeAnnotatable)


class SectionInit(InitBase):

    def __call__(self):
        self.app['schooltool.course.section'] = SectionContainerContainer()


class InstructorsCrowd(Crowd):
    """Crowd of instructors of a section."""

    title = _(u'Instructors')
    description = _(u'Instructors of the section.')

    def contains(self, principal):
        instructors = interfaces.ISection(self.context).instructors
        return IPerson(principal, None) in instructors


class PersonInstructorsCrowd(Crowd):
    """Crowd of instructors of a person."""

    title = _(u'Instructors')
    description = _(u'Instructors of a person in any of his sections.')

    def _getSections(self, ob):
        return [section for section in getRelatedObjects(ob, membership.URIGroup)
                if interfaces.ISection.providedBy(section)]

    def contains(self, principal):
        user = IPerson(principal, None)
        person = IPerson(self.context)
        for section in self._getSections(person):
            if user in section.instructors:
                return True
        return False


class LearnersCrowd(Crowd):
    """Crowd of learners of a section.

    At the moment only direct members of a section are considered as
    learners.
    """
    adapts(interfaces.ISection)

    title = _(u'Learners')
    description = _(u'Students of the section.')

    def contains(self, principal):
        return IPerson(principal, None) in self.context.members


class SectionCalendarSettingCrowd(ConfigurableCrowd):
    adapts(interfaces.ISection)
    setting_key = 'everyone_can_view_section_info'


class SectionEditorSettingCrowd(ConfigurableCrowd):
    adapts(interfaces.ISection)
    setting_key = 'instructors_can_edit_section'

    def contains(self, principal):
        return (ConfigurableCrowd.contains(self, principal) and
                InstructorsCrowd(self.context).contains(principal))


class SectionCalendarViewers(AggregateCrowd):
    """Crowd of those who can see the section calendar."""
    adapts(interfaces.ISection)

    def crowdFactories(self):
        return [ClerksCrowd, AdministratorsCrowd,
                InstructorsCrowd, LearnersCrowd, SectionCalendarSettingCrowd]


class PersonLearnerAdapter(object):
    implements(interfaces.ILearner)
    adapts(IPerson)

    def __init__(self, person):
        self.person = person

    def _getSections(self, ob):
        return [section for section in getRelatedObjects(ob, membership.URIGroup)
                if interfaces.ISection.providedBy(section)]

    def sections(self):
        # First check the the sections a pupil is in directly
        for section in self._getSections(self.person):
            yield section


class PersonInstructorAdapter(object):
    implements(interfaces.IInstructor)
    adapts(IPerson)

    def __init__(self, person):
        self.person = person

    def sections(self):
        return getRelatedObjects(self.person, URISection,
                                 rel_type=URIInstruction)


class RemoveSectionsWhenTermIsDeleted(ObjectEventAdapterSubscriber):
    adapts(IObjectRemovedEvent, ITerm)

    def __call__(self):
        section_container = interfaces.ISectionContainer(self.object)
        for section_id in list(section_container.keys()):
            del section_container[section_id]


class UnlinkSectionWhenDeleted(ObjectEventAdapterSubscriber):
    adapts(IObjectRemovedEvent, interfaces.ISection)

    def __call__(self):
        self.object.previous = None
        self.object.next = None


class SectionLinkContinuinityValidationSubscriber(EventAdapterSubscriber):
    adapts(SectionBeforeLinkingEvent)
    implements(ISubscriber)

    def __call__(self):
        if (self.event.first is None or
            self.event.second is None):
            return # unlinking sections

        first_term = ITerm(self.event.first)
        second_term = ITerm(self.event.second)
        if sameProxiedObjects(first_term, second_term):
            raise InvalidSectionLinkException(
                _("Cannot link sections in same term"))

        if not sameProxiedObjects(ISchoolYear(first_term),
                                  ISchoolYear(second_term)):
            raise InvalidSectionLinkException(
                _("Cannot link sections in different school years"))

        if not sameProxiedObjects(getNextTerm(first_term), second_term):
            raise InvalidSectionLinkException(
                _("Sections must be in consecutive terms"))


def getSectionRosterEventParticipants(event, rel_type):
    if rel_type != event.rel_type:
        return None, None
    if interfaces.ISection.providedBy(event.participant1):
        return event.participant1, event.participant2
    elif interfaces.ISection.providedBy(event.participant2):
        return event.participant2, event.participant1
    else:
        return None, None


def propagateSectionInstructorAdded(event):
    section, teacher = getSectionRosterEventParticipants(event,
        relationships.URIInstruction)
    if section is None:
        return
    if section.next and teacher not in section.next.instructors:
        section.next.instructors.add(teacher)


def propagateSectionInstructorRemoved(event):
    section, teacher = getSectionRosterEventParticipants(event,
        relationships.URIInstruction)
    if section is None:
        return
    if section.next and teacher in section.next.instructors:
        section.next.instructors.remove(teacher)


def propagateSectionStudentAdded(event):
    section, student = getSectionRosterEventParticipants(event,
        relationships.URIMembership)
    if section is None:
        return
    if section.next and student not in section.next.members:
        section.next.members.add(student)


def propagateSectionStudentRemoved(event):
    section, student = getSectionRosterEventParticipants(event,
        relationships.URIMembership)
    if section is None:
        return
    if section.next and student in section.next.members:
        section.next.members.remove(student)


def copySection(section, target_term):
    """Create a copy of a section in a desired term."""
    section_copy = Section(section.title, section.description)
    sections = interfaces.ISectionContainer(target_term)
    name = section.__name__
    if name in sections:
        name = INameChooser(sections).chooseName(name, section_copy)
    sections[name] = section_copy
    for course in section.courses:
        section_copy.courses.add(course)
    for instructor in section.instructors:
        section_copy.instructors.add(instructor)
    for member in section.members:
        section_copy.members.add(member)
    return section_copy


def linkedSectionTermsVocabulary(context):
    section = interfaces.ISection(context)
    linked = section.linked_sections
    terms = [ITerm(s) for s in linked]
    return vocabulary_titled(terms)


def LinkedSectionTermsVocabularyFactory():
    return linkedSectionTermsVocabulary



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Various timetable classes from generation 35.

"""

from persistent import Persistent
from persistent.dict import PersistentDict
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.interface import Interface, implements
from zope.container.interfaces import IContainer, IContained
from zope.container.contained import Contained
from zope.container.btree import BTreeContainer
from zope.location.interfaces import ILocation

from schooltool.app.interfaces import ISchoolToolCalendarEvent
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.calendar.simple import SimpleCalendarEvent


substitutes = {}
def substituteIn(module):
    def register(target):
        name = '%s.%s' % (module, target.__name__)
        substitutes[name] = target
        return target
    return register


@substituteIn('schooltool.timetable.schema')
class TimetableSchemaContainerContainer(BTreeContainer):
    pass


@substituteIn('schooltool.timetable.schema')
class TimetableSchemaContainer(BTreeContainer):
    _default_id = None # default schema for calendars

    default_id = property(lambda self: self._default_id)

    def __delitem__(self, schema_id):
        BTreeContainer.__delitem__(self, schema_id)
        if schema_id == self.default_id:
            self._default_id = None

    def getDefault(self):
        return self[self.default_id]


@substituteIn('schooltool.timetable.interfaces')
class ITimetableSchemaDay(Interface):
    pass


@substituteIn('schooltool.timetable.schema')
class TimetableSchemaDay(Persistent):
    implements(ITimetableSchemaDay)
    periods = None # list of period IDs for this day
    homeroom_period_ids = None # list of homeroom period IDs

    def keys(self):
        return self.periods

    def items(self):
        return [(period, set()) for period in self.periods]

    def __getitem__(self, period):
        # XXX: wow...
        if period not in self.periods:
            raise KeyError(period)
        return set()


@substituteIn('schooltool.timetable.interfaces')
class ITimetableSchema(IContained):
    pass


@substituteIn('schooltool.timetable.schema')
class TimetableSchema(Persistent, Contained):
    implements(ITimetableSchema)

    title = None
    model = None
    timezone = None
    day_ids = None
    days = None

    def keys(self):
        return list(self.day_ids)

    def items(self):
        return [(day, self.days[day]) for day in self.day_ids]

    def __getitem__(self, key):
        return self.days[key]


@substituteIn('schooltool.timetable.interfaces')
class ITimetableModel(Interface):
    pass


@substituteIn('schooltool.timetable.interfaces')
class IWeekdayBasedTimetableModel(ITimetableModel):
    pass


@substituteIn('schooltool.timetable.interfaces')
class IDayIdBasedTimetableModel(ITimetableModel):
    pass


@substituteIn('schooltool.timetable.model')
class WeekdayBasedModelMixin:
    implements(IWeekdayBasedTimetableModel)


@substituteIn('schooltool.timetable.model')
class BaseTimetableModel(Persistent):
    timetableDayIds = () # day ids of templates built by timetable
    dayTemplates = {} # day id -> [(period, SchoolDaySlot), ...]
    exceptionDays = None # date -> [(period, SchoolDaySlot), ...]
    exceptionDayIds = None # exception date -> day id in dayTemplates


@substituteIn('schooltool.timetable.model')
class BaseSequentialTimetableModel(BaseTimetableModel):
    pass


@substituteIn('schooltool.timetable.model')
class WeeklyTimetableModel(BaseTimetableModel,
                           WeekdayBasedModelMixin):
    factory_id = "WeeklyTimetableModel"
    timetableDayIds = "Monday", "Tuesday", "Wednesday", "Thursday", "Friday"


@substituteIn('schooltool.timetable.model')
class SequentialDaysTimetableModel(BaseSequentialTimetableModel,
                                   WeekdayBasedModelMixin):
    factory_id = "SequentialDaysTimetableModel"


@substituteIn('schooltool.timetable.model')
class SequentialDayIdBasedTimetableModel(BaseSequentialTimetableModel):
    implements(IDayIdBasedTimetableModel)
    factory_id = "SequentialDayIdBasedTimetableModel"


@substituteIn('schooltool.timetable')
class SchooldaySlot(object):
    tstart = None
    duration = None

    def __cmp__(self, other):
        return cmp((self.tstart, self.duration),
                   (other.tstart, other.duration))


@substituteIn('schooltool.timetable')
class SchooldayTemplate(object):
    events = ()
    def __iter__(self):
        return iter(sorted(self.events))


@substituteIn('schooltool.timetable.interfaces')
class ITimetableCalendarEvent(ISchoolToolCalendarEvent):
    """A calendar event that has been created from a timetable."""


class STAppCalendarEvent(SimpleCalendarEvent, Persistent, Contained):
    """Taken from schooltool.app.cal.CalendarEvent"""
    implements(ISchoolToolCalendarEvent, IAttributeAnnotatable)


@substituteIn('schooltool.timetable.model')
class PersistentTimetableCalendarEvent(STAppCalendarEvent):
    """A calendar event that has been created from a timetable."""
    implements(ITimetableCalendarEvent)

    __name__ = None
    __parent__ = None # calendar
    unique_id = None

    _day_id = None
    _period_id = None
    _activity = None
    _resources = None

    dtstart = None
    duration = None
    description = None
    location = None
    recurrence = None # XXX: for timetables? really?
    allday = None # XXX: for timetables? really?

    day_id = property(lambda self: self._day_id)
    period_id = property(lambda self: self._period_id)
    activity = property(lambda self: self._activity)
    title = property(lambda self: self.activity.title)
    resources = property(lambda self: self._resources)

    def unbookResource(self, resource):
        if resource not in self.resources:
            raise ValueError('resource not booked')
        self._resources = tuple([r for r in self.resources
                                 if r is not resource])
        ISchoolToolCalendar(resource).removeEvent(self)


@substituteIn('schooltool.timetable.interfaces')
class ITimetableDict(IContainer, ILocation):
    """Container for [section] timetables."""


@substituteIn('schooltool.timetable')
class TimetableDict(PersistentDict):
    """The id of the timetable is composed by joining term id and
    timetable schema id with a dot.  For example,"2005-fall.default"
    means a timetable for a term "2005-fall" with a timetable schema
    "default".
    """
    implements(ILocation, ITimetableDict)

    __name__ = 'timetables'
    __parent__ = None # [section] this timetable is scheduled for


@substituteIn('schooltool.timetable.interfaces')
class ITimetable(ILocation):
    """A timetable interface."""


@substituteIn('schooltool.timetable')
class Timetable(Persistent):
    """A timetable.

    A timetable is an ordered collection of timetable days that contain
    periods. Each period either contains a class, or is empty.

    A timetable represents the repeating lesson schedule for just one
    pupil, or one teacher, or one bookable resource.

    Contained in a TimetableDict.
    """
    implements(ITimetable)

    __name__ = None
    __parent__ = None
    _first = None
    _last = None
    timezone = 'UTC'
    consecutive_periods_as_one = False
    term = None # always set for bound timetables
    schooltt = None # schema, always set for bound timetables

    model = None # model, set from views
    day_ids = None # list of day ids, set from views

    # persistent dict of day templates
    # keys - subset of day_ids
    # values - ITimetableDay, ITimetableDay.timetable set to self
    days = None

    @property
    def title(self):
        if self.term and self.schooltt:
            return "%s.%s" % (self.term.__name__, self.schooltt.__name__)
        else:
            return u"Unbound timetable."

    def __init__(self, day_ids):
        self.day_ids = day_ids
        self.days = PersistentDict()
        self.model = None

    @property
    def first(self):
        if self._first is None:
            term = getattr(self, 'term', None)
            if term is None:
                return None
            return term.first
        return self._first

    @property
    def last(self):
        if self._last is None:
            term = getattr(self, 'term', None)
            if term is None:
                return None
            return term.last
        return self._last

    def keys(self):
        return list(self.day_ids)

    def items(self):
        return [(day, self.days[day]) for day in self.day_ids]

    def __getitem__(self, key):
        return self.days[key]

    def activities(self):
        """Returns a list of tuples (day_id, period_id, activity)."""
        act = []
        for day_id in self.day_ids:
            for period_id, iactivities in self.days[day_id].items():
                act.extend([(day_id, period_id, activity)
                            for activity in iactivities])
        return act


@substituteIn('schooltool.timetable.interfaces')
class ITimetableDay(Interface):
    pass


@substituteIn('schooltool.timetable')
class TimetableDay(Persistent):
    """A day in a timetable.

    A timetable day is an ordered collection of periods that each have
    a set of activites that occur during that period.

    Different days within the same timetable may have different periods.
    """
    implements(ITimetableDay)

    timetable = None
    day_id = None
    periods = None # list of periods (titles, or maybe IDs?)
    homeroom_period_ids = None # IDs of homeroom periods
    activities = None # persistent dict of (period, set of TimetableActivity)

    def keys(self):
        return self.periods

    def items(self):
        return [(period, self.activities[period]) for period in self.periods]

    def __getitem__(self, period):
        return self.activities[period]


@substituteIn('schooltool.timetable.interfaces')
class ITimetableActivity(Interface):
    pass


@substituteIn('schooltool.timetable')
class TimetableActivity(object):
    """Timetable activity.

    Instances are immutable.

    Equivalent timetable activities must compare and hash equally after
    pickling and unpickling.
    """
    implements(ITimetableActivity)

    _title = None # usually course title
    _owner = None # usually section object
    _timetable = None # bound timetable
    _resources = None # tuple of section resources, probably unused nowadays

    title = property(lambda self: self._title)
    owner = property(lambda self: self._owner)
    timetable = property(lambda self: self._timetable)
    resources = property(lambda self: self._resources)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
BasicPerson advisor relationship.
"""
from schooltool.relationship.relationship import RelationshipSchema
from schooltool.relationship.uri import URIObject


URIAdvising = URIObject('http://schooltool.org/ns/advising',
                        'Advising', 'The advising relationship.')
URIStudent = URIObject('http://schooltool.org/ns/advising/student',
                       'Student', 'An advising relationship student role.')
URIAdvisor = URIObject('http://schooltool.org/ns/advising/advisor',
                       'Advisor', 'An advising relationship advisor role.')

Advising = RelationshipSchema(URIAdvising,
                              advisor=URIAdvisor,
                              student=URIStudent)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Sorting algorithms for viewlet managers.
"""
import itertools


def scc_tarjan(items, graph):
    """Tarjan's strongly connected components algorithm."""
    result = []
    stack = []
    low = {}
    lmax = len(items)

    def visit(n):
        if n in low:
            return
        i = len(low)
        low[n] = i
        pos = len(stack)
        stack.append(n)

        edges = graph.get(n, ())
        if edges:
            for m in items:
                if m in edges:
                    visit(m)
                    low[n] = min(low[n], low[m])

        if low[n] == i:
            component = tuple(stack[pos:])
            stack[pos:] = []
            result.append(component)
            for item in component:
                low[item] = lmax

    for n in items:
        visit(n)

    return result


def topological_sort(items, graph, reverse=False):
    L = []
    S = []

    iter_items = lambda: reverse and reversed(items) or iter(items)
    riter_items = lambda: reverse and iter(items) or reversed(items)

    C = dict.fromkeys(items, 0)
    for n in iter_items():
        C[n] = len(graph.get(n, ()))
        if not C[n]:
            S.append(n)

    while S:
        n = S.pop(0)
        if reverse:
            L.insert(0, n)
        else:
            L.append(n)
        for m in riter_items():
            edges = graph.get(m)
            if edges and n in edges:
                C[m] -= 1
                if C[m] == 0:
                    S.insert(0, m)

    if reverse:
        L = [m for m in items if C[m]] + L
    else:
        L += [m for m in items if C[m]]

    return L


def dependency_graph(before, after):
    graph = {}
    for m, ns in before.items():
        for n in ns:
            if n not in graph:
                graph[n] = set()
            graph[n].add(m)

    for n, ms in after.items():
        for m in ms:
            if n not in graph:
                graph[n] = set()
            graph[n].add(m)
    return graph


def scc_presorted(items, before, after):
    presorted = topological_sort(items, before, reverse=True)
    presorted = topological_sort(presorted, after)
    graph = dependency_graph(before, after)
    components = scc_tarjan(presorted, graph)
    return components


def dependency_sort(items, before, after):
    components = scc_presorted(items, before, after)
    return list(itertools.chain(*components))



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2004 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.app.browser.csvimport
"""
import unittest
import doctest

from zope.i18n import translate

from schooltool.app.browser.testing import setUp as testSetUp, tearDown

__metaclass__ = type


def doctest_BaseCSVImporter():
    r"""Test for BaseCSVImporter

    Set up

        >>> from schooltool.app.browser.csvimport import BaseCSVImporter
        >>> importer = BaseCSVImporter(None)

    Subclasses need to define the createAndAdd method

        >>> importer.createAndAdd(None)
        Traceback (most recent call last):
        ...
        NotImplementedError: Please override this method in subclasses

    When given a list of CSV rows parseCSVRows should return a list of lists

        >>> data = ["one, two, three", "four, five, six"]
        >>> importer.parseCSVRows(data)
        [['one', 'two', 'three'], ['four', 'five', 'six']]

    parseCSVRows can also set errors should they occur

        >>> data = ["one, \xff"]
        >>> importer.charset = 'UTF-8'
        >>> importer.parseCSVRows(data)
        >>> translate(importer.errors.generic[0])
        u'Conversion to unicode failed in line 1'

    Moving on, importFromCSV calls createAndAdd which we have not defined

        >>> csvdata = "one, two, three\nfour, five, six"
        >>> importer.importFromCSV(csvdata)
        Traceback (most recent call last):
        ...
        NotImplementedError: Please override this method in subclasses

    We need to make a subclass to test this properly

        >>> class TestCSVImporter(BaseCSVImporter):
        ...     def __init__(self):
        ...         BaseCSVImporter.__init__(self, None, 'UTF-8')
        ...     def createAndAdd(self, data, True):
        ...         pass

        >>> myimporter = TestCSVImporter()

    importFromCSV just returns True if everything goes well

        >>> myimporter.importFromCSV(csvdata)
        True

    False is returned if there are errors

        >>> myimporter.importFromCSV("one, two\nthree, \xff")
        False
        >>> translate(myimporter.errors.generic[0])
        u'Conversion to unicode failed in line 2'

    """


def doctest_ImportErrorCollection():
    r"""
    Make the class

        >>> from schooltool.app.browser.csvimport import ImportErrorCollection
        >>> errors = ImportErrorCollection()
        >>> errors
        <ImportErrorCollection {'generic': [], 'fields': []}>

    anyErrors returns True if there are errors and False if not

        >>> errors.anyErrors()
        False
        >>> errors.fields.append('A Sample Error Message')
        >>> errors.anyErrors()
        True

    """


def setUp(test=None):
    testSetUp(test)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(
        setUp=setUp, tearDown=tearDown,
        optionflags=doctest.ELLIPSIS|doctest.REPORT_NDIFF))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Devmode view support.
"""
from zope.app.exception.browser.unauthorized import Unauthorized
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.viewlet import viewlet

from schooltool.skin import flourish
from schooltool.securitypolicy.policy import CachingSecurityPolicy
from schooltool.app.browser import SchoolToolAPI


DevmodeCSSViewlet = viewlet.CSSViewlet('devmode.css')


class DebugUnauthorized(Unauthorized):
    template = ViewPageTemplateFile('unauthorized.pt')
    next_url = ''

    @property
    def security_checks(self):
        cache = DebugSecurityPolicy.getCache(self.request)
        if cache is None:
            return
        for key, info in cache['debug_order']:
            obj, level = info
            permission = key[0]
            value = cache['perm'].get(key)
            try:
                text = repr(obj)
            except:
                text = '<object>'

            # XXX: quite hacky indeed, but will do for now.
            path = str(getattr(obj, '__name__', ''))
            parent = getattr(obj, '__parent__', None)
            seen = set([id(obj)])
            while parent is not None and id(parent) not in seen:
                seen.add(id(parent))
                path = str(getattr(parent, '__name__', '')) + '/' + path
                parent = getattr(parent, '__parent__', None)

            yield {
                'obj': obj,
                'permission': permission,
                'repr': text,
                'value': value,
                'level': level,
                'name': path,
                }

    def __call__(self):
        cache = DebugSecurityPolicy.getCache(self.request)
        sp_cache_enabled = None
        if cache is not None:
            sp_cache_enabled = cache['enabled']
            cache['enabled'] = False
        result = Unauthorized.__call__(self)
        if self.request.response.getStatus() in (302, 303):
            self.next_url = self.request.response.getHeader('location')
            self.request.response.setHeader('location', '')
            self.request.response.setStatus(200)
            result = self.template()
        if cache is not None:
            cache['enabled'] = sp_cache_enabled
        return result


class FlourishDebugUnauthorized(flourish.page.NoSidebarPage,
                                DebugUnauthorized):
    container_class = 'container extra-wide-container'

    def render(self, *args, **kw):
        return DebugUnauthorized.__call__(self)


class DebugSecurityPolicy(CachingSecurityPolicy):

    @classmethod
    def getCache(cls, participation):
        cache = CachingSecurityPolicy.getCache(participation)
        if (cache is not None and
            'debug_order' not in cache):
            cache['debug_order'] = []
            cache['debug_level'] = 0
        return cache

    def checkPermission(self, permission, obj):
        caches = filter(None, [self.getCache(p) for p in self.participations])
        levels = [cache['debug_level'] for cache in caches]
        perm = CachingSecurityPolicy.checkPermission(
            self, permission, obj)
        for cache, level in zip(caches, levels):
            cache['debug_level'] = level
        return perm

    def cache(self, participation, permission, obj, value):
        cache = self.getCache(participation)
        if cache is None or not cache['enabled']:
            return
        cache['debug_level'] += 1
        CachingSecurityPolicy.cache(
            self, participation, permission, obj, value)
        key = self.cachingKey(permission, obj)
        if (key is not None and
            key in cache['perm'] and
            key not in cache['debug_order']):
            extra_info = (obj, cache['debug_level'])
            cache['debug_order'].append((key, extra_info))


class DevmodeSchoolToolAPI(SchoolToolAPI):

    devmode = True


def launch_pdb_on_exception():
    import zope.publisher.publish
    import sys
    try:
        import ipdb as pdb
    except ImportError:
        import pdb
    olde_debug_call = zope.publisher.publish.debug_call
    def interceptor(*args, **kw):
        try:
            result = olde_debug_call(*args, **kw)
        except:
            type, value, tb = sys.exc_info()
            pdb.post_mortem(tb)
            raise
        return result
    zope.publisher.publish.debug_call = interceptor

#launch_pdb_on_exception()



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Catalog indexing extensions for tabling."""

from persistent import Persistent

from zope.interface import implements, implementsOnly
from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.container.contained import Contained
from zope.catalog.interfaces import ICatalogIndex
from zope.catalog.interfaces import ICatalog
from zope.intid.interfaces import IIntIds

from zc.catalog.index import SetIndex, ValueIndex
from zc.catalog.interfaces import IValueIndex, ISetIndex
from zc.catalog.interfaces import IExtentCatalog

from schooltool.table.interfaces import IIndexedTableFormatter
from schooltool.table.interfaces import IIndexedColumn
from schooltool.table.table import FilterWidget
from schooltool.table.table import SchoolToolTableFormatter
from schooltool.table.table import url_cell_formatter

# BBB: imports
from schooltool.table.column import IndexedGetterColumn
from schooltool.table.column import IndexedLocaleAwareGetterColumn
from schooltool.table.column import makeIndexedColumn
from schooltool.table.column import RenderUnindexingMixin, unindex

from schooltool.common import SchoolToolMessage as _


class FilterImplementing(Persistent):
    def __init__(self, interface):
        self.interface = interface

    def __call__(self, index, docid, item):
        return bool(self.interface.providedBy(item))


class ConvertingIndexMixin(object):
    def __init__(self, converter=None, *args, **kwargs):
        assert converter is not None
        super(ConvertingIndexMixin, self).__init__(*args, **kwargs)
        self.value_factory = converter

    def index_doc(self, docid, texts):
        value = self.value_factory(texts)
        super(ConvertingIndexMixin, self).index_doc(docid, value)


class IConvertingIndex(IValueIndex, ICatalogIndex):
    """Index of values created by external converter."""


class ConvertingIndex(ConvertingIndexMixin, ValueIndex, Contained):
    implements(IConvertingIndex)


class ConvertingSetIndexMixin(object):
    def __init__(self, converter=None, *args, **kwargs):
        assert converter is not None
        super(ConvertingSetIndexMixin, self).__init__(*args, **kwargs)
        self.value_factory = converter

    def index_doc(self, docid, texts):
        value = self.value_factory(texts)
        super(ConvertingSetIndexMixin, self).index_doc(docid, value)


class IConvertingSetIndex(ISetIndex, ICatalogIndex):
    """Index of values created by external converter."""


class ConvertingSetIndex(ConvertingSetIndexMixin, SetIndex, Contained):
    implements(IConvertingSetIndex)


class IndexedFilterWidget(FilterWidget):

    search_index = 'title'

    @Lazy
    def catalog(self):
        return ICatalog(self.source)

    def filter(self, items):
        index = self.catalog[self.search_index]
        if 'SEARCH' in self.request and 'CLEAR_SEARCH' not in self.request:
            searchstr = self.request['SEARCH'].lower()
            results = []
            for item in items:
                title = index.documents_to_values[item['id']]
                if searchstr in title.lower():
                    results.append(item)
        else:
            self.request.form['SEARCH'] = ''
            results = items

        return results


class IndexedTableFormatter(SchoolToolTableFormatter):
    implementsOnly(IIndexedTableFormatter)

    def columns(self):
        return [IndexedGetterColumn(name='title',
                                    title=_(u"Title"),
                                    getter=lambda i, f: i.title,
                                    cell_formatter=url_cell_formatter,
                                    index='title')]

    @Lazy
    def catalog(self):
        return ICatalog(self.source)

    def items(self):
        """Return a list of index dicts for all the items in the context container"""
        catalog = self.catalog
        if IExtentCatalog.providedBy(catalog):
            ids = list(catalog.extent)
        else:
            index = catalog.values()[0]
            ids = index.documents_to_values.keys()

        items = [{'id': id, 'catalog': catalog}
                 for id in ids]
        return items

    def ommit(self, items, ommited_items):
        if not ommited_items:
            return items
        ommited_items = self.indexItems(ommited_items)
        ommited_ids = set([item['id'] for item in ommited_items])
        return [item for item in items
                if item['id'] not in ommited_ids]

    def indexItems(self, items):
        """Convert a list of objects to a list of index dicts"""
        int_ids = getUtility(IIntIds)
        catalog = self.catalog
        results = []
        for item in items:
            results.append({
                    'id': int_ids.getId(item),
                    'catalog': catalog})
        return results

    def getItem(self, indexed):
        int_ids = getUtility(IIntIds)
        return int_ids.queryObject(indexed['id'])

    def setUp(self, **kwargs):
        items = kwargs.pop('items', None)
        if items is not None:
            items = self.indexItems(items)
        super(IndexedTableFormatter, self).setUp(
            items=items, **kwargs)

    def makeFormatter(self):
        if self._table_formatter is None:
            return None
        columns = [IIndexedColumn(c) for c in self._columns]
        formatter = self._table_formatter(
            self.source, self.request, self._items,
            columns=columns,
            batch_start=self.batch.start, batch_size=self.batch.size,
            sort_on=self._sort_on,
            prefix=self.prefix)
        formatter.cssClasses.update(dict(self.css_classes))
        return formatter

